#!/usr/bin/env python3
# Install aws cli and run aws configure to input credentials first and then:
# python mbtiles2s3.py ../data/administrative.mbtiles s3://ss-vector-tile-data/sovereign/v20240410/administrative -p

import json
import logging
import os
import sqlite3
from functools import partial
from multiprocessing.pool import ThreadPool
from urllib.parse import urlparse

import boto3
import click

# import utils

upload_progress_interval = 100
tile_count = 0
upload_count = 0


class MBTilesGenerator(object):
    """Generator that returns tiles from an mbtiles file"""

    def __init__(self, mbtiles):
        super(MBTilesGenerator, self).__init__()
        self.db = sqlite3.connect(mbtiles)
        self.cursor = self.db.cursor()
        self.cursor.execute(
            "SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles order by zoom_level"
        )

    def len(self):
        c = self.db.cursor()
        c.execute("SELECT count(1) from tiles")
        return c.fetchone()[0]

    def __iter__(self):
        return self

    def __next__(self):
        row = self.cursor.fetchone()
        if row == None:
            raise StopIteration()
        zoom, x, y, tile = row
        y = ((1 << zoom) - y) - 1
        return zoom, x, y, tile


def get_tile_json(mbtiles, bucket, key_template):
    db = sqlite3.connect(mbtiles)
    cursor = db.cursor()
    cursor.execute("SELECT name, value FROM metadata")
    tilejson = {
        "tilejson": "2.2.0",
        "scheme": "xyz",
        "tiles": ["https://s3.amazonaws.com/{}/{}".format(bucket, key_template)],
    }
    for key, value in cursor.fetchall():
        if key == "json":
            data = json.loads(value)
            tilejson.update(data)
        else:
            if key in ("center", "bounds"):
                value = [float(s) for s in value.split(",")]
            elif key in ("minzoom", "maxzoom"):
                value = int(value)
            tilejson[key] = value
    return tilejson


def upload_tile(
    s3, bucket, key_template, headers, tile_stuff, progress=True, retries=0
):
    try:
        zoom, x, y, tile = tile_stuff
        s3.put_object(
            Body=tile,
            Bucket=bucket,
            Key=key_template.format(z=zoom, x=x, y=y),
            ContentType=headers.get("Content-Type", ""),
            ContentEncoding=headers.get("Content-Encoding", ""),
            CacheControl=headers.get("Cache-Control", ""),
        )
        global upload_count
        upload_count += 1
        if progress and upload_count % upload_progress_interval == 0:
            print("%i/%i" % (upload_count, tile_count))
    except Exception as e:
        logging.error(str(e))
        if retries < 2:
            upload_tile(
                s3,
                bucket,
                key_template,
                headers,
                tile_stuff,
                progress=progress,
                retries=retries + 1,
            )
        else:
            raise Exception("Too Many upload failures")


@click.command()
@click.argument("mbtiles", type=click.Path(exists=True), required=True)
@click.argument("s3_url", required=True)
@click.option("--threads", default=10, help="Number of simultaneous uploads")
@click.option("--extension", default=".pbf", help="File extension for tiles")
@click.option("--header", "-h", multiple=True, help="Additional headers")
@click.option(
    "--progress", "-p", default=False, is_flag=True, help="Show upload progress"
)
@click.option("--debug", "-d", default=False, help="Debug level logging", is_flag=True)
def upload(mbtiles, s3_url, threads, extension, header, progress, debug):
    """Upload tiles from an MBTiles file to S3.

    \b
    PARAMS:
        mbtiles: Path to an MBTiles file
        s3_url: url to an s3 bucket to upload tiles to
    """
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)
    logging.getLogger("botocore.credentials").setLevel(logging.getLevelName("ERROR"))
    logging.getLogger(
        "botocore.vendored.requests.packages.urllib3.connectionpool"
    ).setLevel(logging.getLevelName("ERROR"))
    logging.getLogger("urllib3.connectionpool").setLevel(logging.getLevelName("ERROR"))

    base_url = urlparse(s3_url)

    s3 = boto3.client("s3")
    bucket = base_url.netloc
    key_prefix = base_url.path.lstrip("/")

    headers = {}
    if header is not None:
        for h in header:
            k, v = h.split(":")
            if k not in ("Cache-Control", "Content-Type", "Content-Encoding"):
                raise Exception("Unsupported header")
            headers[k] = v

    if extension == ".pbf":
        headers.update(
            {
                "Content-Encoding": "gzip",
                "Content-Type": "application/vnd.mapbox-vector-tile",
                "Content-Type": "application/x-protobuf",
            }
        )
    elif extension == ".webp":
        headers.update({"Content-Type": "image/webp"})
    elif extension == ".png":
        headers.update({"Content-Type": "image/png"})
    elif extension == ".jpg" or extension == ".jpeg":
        headers.update({"Content-Type": "image/jpeg"})

    tiles = MBTilesGenerator(mbtiles)
    global tile_count
    tile_count = tiles.len()

    key_template = key_prefix + "/{z}/{x}/{y}" + extension
    logging.info(f"uploading tiles from {mbtiles} to s3://{bucket}/{key_template}")
    pool = ThreadPool(threads)
    func = partial(upload_tile, s3, bucket, key_template, headers, progress=progress)
    pool.map(func, tiles)

    tilejson_key = "{}/index.json".format(key_prefix.strip("/"))
    logging.info(f"uploading tilejson to s3://{bucket}/{tilejson_key}")
    tilejson_data = get_tile_json(mbtiles, bucket, key_template)
    tilejson_json = json.dumps(tilejson_data, indent=4, sort_keys=True)
    s3.put_object(
        Body=json.dumps(tilejson_data),
        Bucket=bucket,
        Key=tilejson_key,
        ContentType="application/json",
    )


if __name__ == "__main__":
    upload()



from collections import namedtuple

from .Mapbox import vector_tile_pb2 as vector_tile
from .utils import CMD_LINE_TO, CMD_MOVE_TO, LINESTRING, zig_zag_decode, zig_zag_encode


class StringTableOptimiser:
    """
    Optimizes the order of keys and values in the MVT layer string table.

    Counts the number of times an entry in the MVT string table (both keys and values) is used. Then reorders the
    string table to have the most commonly used entries first and updates the features to use the replacement
    locations in the table. This can save several percent in a tile with large numbers of features.
    """

    def __init__(self):
        self.key_counts = {}
        self.val_counts = {}

    def add_tags(self, feature_tags):
        itr = iter(feature_tags)
        for k, v in zip(itr, itr):
            self.key_counts[k] = self.key_counts.get(k, 0) + 1
            self.val_counts[v] = self.val_counts.get(v, 0) + 1

    @staticmethod
    def _update_table(counts, table):
        # Sort string table by usage, so most commonly-used values are
        # assigned the smallest indices. Since indices are encoded as
        # varints, this should make best use of the space.
        sort = sorted(((c, k) for k, c in counts.items()), reverse=True)

        # construct the re-ordered string table
        new_table = []
        for _, x in sort:
            new_table.append(table[x])
        assert len(new_table) == len(table)

        # delete table in place and replace with the new table
        del table[:]
        table.extend(new_table)

        # construct a lookup table from the old to the new indices.
        new_indexes = {}
        for i, (_, k) in enumerate(sort):
            new_indexes[k] = i

        return new_indexes

    def update_string_table(self, layer):
        new_key = self._update_table(self.key_counts, layer.keys)
        new_val = self._update_table(self.val_counts, layer.values)

        for feature in layer.features:
            for i in range(0, len(feature.tags), 2):
                feature.tags[i] = new_key[feature.tags[i]]
                feature.tags[i + 1] = new_val[feature.tags[i + 1]]


# We assume that every linestring consists of a single MoveTo command followed by some number of LineTo commands,
# and we encode this as a Line object.
#
# Normally, MVT linestrings are encoded relative to the preceding linestring (if any) in the geometry. However,
# that's awkward for reordering, so we construct an absolute MoveTo for each Line. We also derive a corresponding
# EndsAt location, which isn't used in the encoding, but simplifies analysis.
MoveTo = namedtuple("MoveTo", "x y")
EndsAt = namedtuple("EndsAt", "x y")
Line = namedtuple("Line", "moveto endsat cmds")


def _decode_lines(geom):
    """
    Decode a linear MVT geometry into a list of Lines.

    Each individual linestring in the MVT is extracted to a separate entry in the list of lines.
    """

    lines = []
    current_line = []
    current_moveto = None

    # To keep track of the position. We'll adapt the move-to commands to all be relative to 0,0 at the beginning of
    # each linestring.
    x = 0
    y = 0

    end = len(geom)
    i = 0
    while i < end:
        header = geom[i]
        cmd = header & 7
        run_length = header // 8

        if cmd == CMD_MOVE_TO:
            # flush previous line.
            if current_moveto:
                lines.append(Line(current_moveto, EndsAt(x, y), current_line))
                current_line = []

            assert run_length == 1
            x += zig_zag_decode(geom[i + 1])
            y += zig_zag_decode(geom[i + 2])
            i += 3

            current_moveto = MoveTo(x, y)

        elif cmd == CMD_LINE_TO:
            assert current_moveto

            # we just copy this run, since it's encoding isn't going to change
            next_i = i + 1 + run_length * 2
            current_line.extend(geom[i:next_i])

            # but we still need to decode it to figure out where each move-to command is in absolute space.
            for j in range(run_length):
                dx = zig_zag_decode(geom[i + 1 + 2 * j])
                dy = zig_zag_decode(geom[i + 2 + 2 * j])
                x += dx
                y += dy

            i = next_i

        else:
            raise ValueError(f"Unhandled command: {cmd}")

    if current_line:
        assert current_moveto
        lines.append(Line(current_moveto, EndsAt(x, y), current_line))

    return lines


def _reorder_lines(lines):
    """
    Reorder lines so that the distance from the end of one to the beginning of the next is minimized.
    """

    x = 0
    y = 0
    new_lines = []

    # treat the list of lines as a stack, off which we keep popping the best one to add next.
    while lines:
        # looping over all the lines like this isn't terribly efficient, but in local tests seems to handle a few
        # thousand lines without a problem.
        min_dist = None
        min_i = None
        for i, line in enumerate(lines):
            moveto, _, _ = line

            dist = abs(moveto.x - x) + abs(moveto.y - y)
            if min_dist is None or dist < min_dist:
                min_dist = dist
                min_i = i

        assert min_i is not None
        line = lines.pop(min_i)
        _, endsat, _ = line
        x = endsat.x
        y = endsat.y
        new_lines.append(line)

    return new_lines


def _rewrite_geometry(geom, new_lines):
    """
    Re-encode a list of Lines with absolute MoveTos as a continuous stream of MVT geometry commands, each relative to
    the last. Replace geom with that stream.
    """

    new_geom = []
    x = 0
    y = 0
    for line in new_lines:
        moveto, endsat, lineto_cmds = line

        dx = moveto.x - x
        dy = moveto.y - y
        x = endsat.x
        y = endsat.y

        new_geom.append(9)  # move to, run_length = 1
        new_geom.append(zig_zag_encode(dx))
        new_geom.append(zig_zag_encode(dy))
        new_geom.extend(lineto_cmds)

    # write the lines back out to geom
    del geom[:]
    geom.extend(new_geom)


def optimise_multilinestring(geom):
    # Split the geometry into multiple lists, each starting with a move-to command and consisting otherwise of
    # line-to commands. (perhaps with a close at the end? Is that allowed for linestrings?)

    lines = _decode_lines(geom)

    # can't reorder anything unless it has multiple lines.
    if len(lines) > 1:
        lines = _reorder_lines(lines)
        _rewrite_geometry(geom, lines)


def optimise_tile(tile_bytes):
    """
    Decode a sequence of bytes as an MVT tile and reorder the string table of its layers and the order of its
    multilinestrings to save a few bytes.
    """

    t = vector_tile.tile()
    t.ParseFromString(tile_bytes)

    for layer in t.layers:
        sto = StringTableOptimiser()

        for feature in layer.features:
            # (multi)linestrings only
            if feature.type == LINESTRING:
                optimise_multilinestring(feature.geometry)

            sto.add_tags(feature.tags)

        sto.update_string_table(layer)

    return t.SerializeToString()


if __name__ == "__main__":
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="Input MVT file", type=argparse.FileType("r"))
    parser.add_argument(
        "--output-file", help="Output file, default is stdout", type=argparse.FileType("w"), default=sys.stdout
    )
    args = parser.parse_args()

    output_bytes = optimise_tile(args.input_file.read())
    args.output_file.write(output_bytes)



import json
import argparse
import sqlite3
from vtiles.utils.vt2geojson.tools import vt_bytes_to_geojson
import gzip
import zlib
import logging
from tqdm import tqdm

def tile_data_to_geojson(tile_data, x, y, z, layers):
    """
    Convert tile data to GeoJSON format, filtering by specified layers, and add zoom level to each feature's properties.

    Args:
        tile_data (bytes): The binary tile data.
        x (int): The tile column.
        y (int): The tile row.
        z (int): The zoom level of the tile.
        layers (list): List of layer names to include in the output.

    Returns:
        dict: The filtered GeoJSON features with zoom level included.
    """
    try:
        features = vt_bytes_to_geojson(tile_data, x, y, z)
        if layers:
            filtered_features = {layer: features[layer] for layer in layers if layer in features}
        else:
            filtered_features = features

        # Add zoom level to each feature's properties
        for layer, feature_collection in filtered_features.items():
            for feature in feature_collection['features']:
                if 'properties' not in feature:
                    feature['properties'] = {}
                feature['properties']['zoom_level'] = z
        
        return filtered_features
    except Exception as e:
        logging.error(f"Error converting tile data to GeoJSON at tile ({x}, {y}, {z}): {e}")
        return None

def decompress_tile_data(tile_data):
    """
    Decompress tile data if it is compressed.

    Args:
        tile_data (bytes): The binary tile data.

    Returns:
        bytes: The decompressed tile data.
    """
    try:
        if tile_data[:2] == b'\x1f\x8b':  # Check for gzip magic number
            return gzip.decompress(tile_data)
        elif tile_data[:2] in {b'\x78\x9c', b'\x78\x01', b'\x78\xda'}:
            return zlib.decompress(tile_data)
        return tile_data
    except Exception as e:
        logging.error(f"Failed to decompress tile data: {e}")
        return tile_data
    
def merge_geojsons(geojson_list):
    """
    Merge a list of GeoJSON feature collections into a single feature collection.

    Args:
        geojson_list (list): List of GeoJSON feature collections.

    Returns:
        dict: Merged GeoJSON feature collection.
    """
    merged_geojson = {}

    for geojson in geojson_list:
        for key, feature_collection in geojson.items():
            if key not in merged_geojson:
                merged_geojson[key] = {'type': 'FeatureCollection', 'features': []}
            merged_geojson[key]['features'].extend(feature_collection['features'])
    
    return merged_geojson

def mbtiles_to_geojson(input_mbtiles, output_geojson, min_zoom, max_zoom, flip_y, layers):
    """
    Convert MBTiles data to GeoJSON format.

    Args:
        input_mbtiles (str): Path to the input MBTiles file.
        output_geojson (str): Path to the output GeoJSON file.
        min_zoom (int): Minimum zoom level to extract.
        max_zoom (int): Maximum zoom level to extract.
        flip_y (bool): Whether to flip the y coordinate (TMS format).
        layers (list): List of layer names to include in the output.
    """
    all_features = []
    try:
        conn = sqlite3.connect(input_mbtiles)
        cursor = conn.cursor()

        for zoom_level in range(min_zoom, max_zoom + 1):
            cursor.execute('''
                SELECT tile_column, tile_row, tile_data 
                FROM tiles 
                WHERE zoom_level=?''', (zoom_level,))
            rows = cursor.fetchall()

            for x, y, tile_data in tqdm(rows, desc=f"Converting tiles at zoom level {zoom_level} to GeoJSON"):
                if flip_y:
                    y = (1 << zoom_level) - 1 - y
                tile_data = decompress_tile_data(tile_data)
                if tile_data:
                    features = tile_data_to_geojson(tile_data, x, y, zoom_level, layers)
                    if features:
                        all_features.append(features)

        conn.close()

        merged_geojson = merge_geojsons(all_features)
        with open(output_geojson, 'w') as f:
            json.dump(merged_geojson, f, indent=2)
        logging.info(f"GeoJSON data has been saved to {output_geojson}")

    except sqlite3.Error as e:
        logging.error(f"Failed to read MBTiles file {input_mbtiles}: {e}")
    except Exception as e:
        logging.error(f"Failed to convert {input_mbtiles} to GeoJSON: {e}")

def main():
    """
    Main function to parse arguments and convert MBTiles to GeoJSON.
    """
    parser = argparse.ArgumentParser(description='Convert MBTiles to GeoJSON.')
    parser.add_argument('-i', '--input', type=str, required=True, help='Input MBTiles file')
    parser.add_argument('-o', '--output', type=str, required=True, help='Output GeoJSON file')
    parser.add_argument('-minzoom','--minzoom', type=int, required=True, help='Minimum tile zoom level')
    parser.add_argument('-maxzoom','--maxzoom', type=int, required=True, help='Maximum tile zoom level')
    parser.add_argument('-flipy', '--flipy', type=int, choices=[0, 1], default=0, help='TMS <--> XYZ tiling scheme (optional): 1 or 0, default is 0')
    parser.add_argument('-l', '--layers', type=str, nargs='*', help='List of layer names to convert')

    args = parser.parse_args()
    
    mbtiles_to_geojson(args.input, args.output, args.minzoom, args.maxzoom, args.flipy == 1, args.layers)

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    main()



import sqlite3
import shutil
import gzip, zlib
import json
import argparse
import os
from tqdm import tqdm
from vtiles.utils.mapbox_vector_tile import encode, decode
from vtiles.utils.geopreocessing import fix_wkt
import logging
from vtiles.mbtiles import mbtilesfixmeta

logging.basicConfig(level=logging.INFO)

def process_metadata(metadata_json, layers_to_keep, exclude=False):
    vector_layers = metadata_json.get('vector_layers', [])
    if exclude:
        processed_layers = [layer for layer in vector_layers if layer['id'] not in layers_to_keep]
    else:
        processed_layers = [layer for layer in vector_layers if layer['id'] in layers_to_keep]
    
    tilestats = metadata_json.get('tilestats', {})
    updated_stats = tilestats.copy()
    updated_stats['layerCount'] = len(processed_layers)
    updated_stats['layers'] = [
                            stat for stat in updated_stats.get('layers', [])
                            if (exclude and stat['layer'] not in layers_to_keep) or (not exclude and stat['layer'] in layers_to_keep)
                            ]

    metadata_json['vector_layers'] = processed_layers
    metadata_json['tilestats'] = updated_stats
    
    return metadata_json


def process_mbtiles(input_mbtiles, output_mbtiles, layers_to_keep, keep_layers=True):
    if os.path.exists(output_mbtiles):
        os.remove(output_mbtiles)
    is_vector, compression_type = mbtilesfixmeta.check_vector(input_mbtiles) 
    if is_vector:
        mbtilesfixmeta.fix_metadata(input_mbtiles, compression_type)    
        shutil.copyfile(input_mbtiles, output_mbtiles)    
        with sqlite3.connect(output_mbtiles) as conn:
            cursor = conn.cursor()
            
            cursor.execute("SELECT name, value FROM metadata WHERE name = 'json'")
            metadata = cursor.fetchone()
            
            if metadata:
                _, metadata_value = metadata
                metadata_json = json.loads(metadata_value)
                processed_metadata = process_metadata(metadata_json, layers_to_keep, exclude=not keep_layers)
                cursor.execute("DELETE FROM metadata WHERE name = 'json'")
                cursor.execute("INSERT INTO metadata (name, value) VALUES ('json', ?)", (json.dumps(processed_metadata),))
            
            cursor.execute("SELECT type FROM sqlite_master WHERE name='tiles'")
            result = cursor.fetchone()
            if result:
                if result[0] == 'view':
                    cursor.execute("DROP VIEW IF EXISTS tiles")
                else:
                    cursor.execute("DROP TABLE IF EXISTS tiles")
            
            cursor.execute("""
                CREATE TABLE tiles (
                    zoom_level INTEGER,
                    tile_column INTEGER,
                    tile_row INTEGER,
                    tile_data BLOB
                )
            """)
            cursor.execute("CREATE UNIQUE INDEX tile_index ON tiles (zoom_level, tile_column, tile_row)")
            
            try:
                with sqlite3.connect(input_mbtiles) as in_conn:
                    in_cursor = in_conn.cursor()
                    in_cursor.execute("SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles ORDER BY zoom_level")
                    tiles = in_cursor.fetchall()
                    
                    for zoom_level, tile_column, tile_row, tile_data in tqdm(tiles, desc="Processing tiles", unit=" tiles"):
                        if tile_data[:2] == b'\x1f\x8b':
                            tile_data = gzip.decompress(tile_data)
                        elif tile_data[:2] in [b'\x78\x9c', b'\x78\x01', b'\x78\xda']:
                            tile_data = zlib.decompress(tile_data)
                        
                        decoded_tile = decode(tile_data)
                        decoded_tile = fix_wkt(decoded_tile)
                        
                        if keep_layers:
                            filtered_tile = [item for item in decoded_tile if item["name"] in layers_to_keep]
                        else:
                            filtered_tile = [item for item in decoded_tile if item["name"] not in layers_to_keep]
                        
                        if filtered_tile:
                            try:
                                encoded_tile = encode(filtered_tile)
                                encoded_tile_gzip = gzip.compress(encoded_tile)
                                cursor.execute("""
                                    INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data)
                                    VALUES (?, ?, ?, ?)""", (zoom_level, tile_column, tile_row, encoded_tile_gzip))
                            except Exception as e:
                                logging.error(f"Error encoding tile {zoom_level}/{tile_column}/{tile_row}: {e}")
                
                description = 'Splitting MBTiles file by selected layers using mbtilessplit from vtiles'
                cursor.execute('''
                    INSERT OR REPLACE INTO metadata (name, value)
                    VALUES ('description', ?)
                ''', (description,))
                
                conn.commit()
                if keep_layers:
                    logging.info(f'Successfully saved split MBTiles into {output_mbtiles}')
                else:
                    logging.info(f'Successfully saved remaining MBTiles into {output_mbtiles}')
            
            except sqlite3.Error as e:
                logging.error(f"Database error: {e}")
            except Exception as e:
                logging.error(f"Unexpected error: {e}")


def main():
    parser = argparse.ArgumentParser(description="Split an MBTiles file by selected layers.")
    parser.add_argument("-i", "--input", required=True, help="Path to the input MBTiles file")
    parser.add_argument("-o", "--output", required=True, help="Path to the output MBTiles file")
    parser.add_argument("-l", "--layers", nargs='+', required=True, help="List of layer names to be splitted")
    args = parser.parse_args()
    
    process_mbtiles(args.input, args.output, args.layers, keep_layers=True)
    remaining_output = f"{os.path.splitext(args.input)[0]}_remained.mbtiles"
    process_mbtiles(args.input, remaining_output, args.layers, keep_layers=False)

if __name__ == "__main__":
    main()



from vtiles.utils.mapbox_vector_tile import decode
import sys
import gzip
import zlib
import os
from datetime import datetime

def read_pbf_tile(pbf_file):
    try:
        with open(pbf_file, 'rb') as f:
            tile_data = f.read()

        # Check for GZIP compression
        if tile_data[:2] == b'\x1f\x8b':
            compression_type = 'GZIP'
            tile_data = gzip.decompress(tile_data)
        # Check for ZLIB compression
        elif tile_data[:2] == b'\x78\x9c' or tile_data[:2] == b'\x78\x01' or tile_data[:2] == b'\x78\xda':
            compression_type = 'ZLIB'
            tile_data = zlib.decompress(tile_data)
        else:
            compression_type = 'None'

        # Decode the tile data
        tile = decode(tile_data)
        return tile, compression_type

    except Exception as e:
        print(f"Error reading or decoding the PBF file: {e}")
        sys.exit(1)

def count_feature_types(layer_data):
    from collections import defaultdict
    feature_types_count = defaultdict(int)  # Use defaultdict to count feature types
    for feature in layer_data['features']:
        geometry_type = feature['geometry']['type']
        feature_types_count[geometry_type] += 1
    return feature_types_count

def main():
    if len(sys.argv) < 2:
        print("Usage: pbfinfo <path_to_pbf_file>")
        sys.exit(1)

    pbf_file = sys.argv[1]

    # Get file size and last modified date
    file_size = os.path.getsize(pbf_file)
    last_modified_timestamp = os.path.getmtime(pbf_file)
    last_modified_date = datetime.fromtimestamp(last_modified_timestamp).strftime('%Y-%m-%d %H:%M:%S')

    # Read and parse the PBF tile
    tile_data, compression_type = read_pbf_tile(pbf_file)

    print(f"File size: {file_size} bytes")
    print(f"Last modified: {last_modified_date}")
    print(f"Compression type: {compression_type}")
    # Print layer information
    print("==============")
    for layer_name, layer_data in tile_data.items():
        num_features = len(layer_data['features'])
        feature_types_count = count_feature_types(layer_data)
        
        print(f"Layer '{layer_name}':")
        print(f"  Total features: {num_features} features")
        print("  Feature types:")
        for feature_type, count in feature_types_count.items():
            print(f"    {feature_type}: {count} features")

if __name__ == "__main__":
    main()


import sys
import json
import osmium as o
from tqdm import tqdm  # Import tqdm for progress bar
import os 

geojsonfab = o.geom.GeoJSONFactory()

class GeoJsonWriter(o.SimpleHandler):
    def __init__(self, output_file):
        super().__init__()
        self.output_file = output_file
        self.features = []
    
    def finish(self):
        try:
            with open(self.output_file, 'w') as f:
                f.write('{"type": "FeatureCollection", "features": [\n')
                for i, feature in enumerate(tqdm(self.features, desc='Writing GeoJSON')):
                    if i > 0:
                        f.write(',\n')
                    json.dump(feature, f)
                f.write('\n]}')
            print(f"GeoJSON file '{self.output_file}' successfully created.")
        except Exception as e:
            print(f"Error: Failed to write GeoJSON file '{self.output_file}': {e}")

    def node(self, n):
        if n.tags:
            self.add_feature(geojsonfab.create_point(n), n.tags)

    def way(self, w):
        if w.tags and not w.is_closed():
            self.add_feature(geojsonfab.create_linestring(w), w.tags)

    def area(self, a):
        if a.tags:
            self.add_feature(geojsonfab.create_multipolygon(a), a.tags)

    def add_feature(self, geojson, tags):
        geom = json.loads(geojson)
        if geom:
            feature = {'type': 'Feature', 'geometry': geom, 'properties': dict(tags)}
            self.features.append(feature)

def main():
    if len(sys.argv) != 5 or sys.argv[1] != '-i' or sys.argv[3] != '-o':
        print(f"Usage: {sys.argv[0]} -i <path_to_osm_file.osm.pbf> -o <output_file.geojson>")
        sys.exit(1)
    
    osmfile = sys.argv[2]
    output_file = sys.argv[4]
    
    if not os.path.isfile(osmfile):
        print(f"Error: File {osmfile} not found.")
        sys.exit(1)

    handler = GeoJsonWriter(output_file)
    handler.apply_file(osmfile)
    handler.finish()

if __name__ == '__main__':
    main()



from http.server import SimpleHTTPRequestHandler, HTTPServer
import os

class CustomHTTPRequestHandler(SimpleHTTPRequestHandler):

    def check_compressed(self, pbf_file):
        try:
            with open(pbf_file, 'rb') as f:
                tile_data = f.read(2)  # Read only the first two bytes
            # Check for GZIP compression
            if tile_data[:2] == b'\x1f\x8b':
                return 'GZIP'
            # Check for ZLIB compression
            elif tile_data[:2] == b'\x78\x9c' or tile_data[:2] == b'\x78\x01' or tile_data[:2] == b'\x78\xda':
                return 'ZLIB'
            else:
                return None
        except Exception as e:
            print(f"Error reading PBF file: {e}")
            return None

    def end_headers(self):
        # If the requested file is a .pbf file
        if self.path.endswith('.pbf'):
            file_path = self.translate_path(self.path)  # Get the full file path
            compression_type = self.check_compressed(file_path)  # Check the compression type

            if compression_type == 'GZIP':
                self.send_header('Content-Encoding', 'gzip')
            elif compression_type == 'ZLIB':
                self.send_header('Content-Encoding', 'deflate')  # ZLIB uses 'deflate' as encoding

        SimpleHTTPRequestHandler.end_headers(self)

    def guess_type(self, path):
        base, ext = os.path.splitext(path)
        if ext.lower() == '.pbf':
            return 'application/x-protobuf'
        else:
            return SimpleHTTPRequestHandler.guess_type(self, path)

def main():
    port = 8000  # Set the port
    server_address = ('', port)  # Set the server address
    handler_class = CustomHTTPRequestHandler  # Specify the custom request handler
    httpd = HTTPServer(server_address, handler_class)  # Create the HTTP server
    print(f"Simple Vector Tile Server for tiles cache folder running on port {port}")
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        print("Keyboard interrupt received, stopping server...")
        httpd.server_close()

if __name__ == "__main__":
    main()



import sqlite3, argparse, sys, math, logging, time, os, json
from tqdm import tqdm
logger = logging.getLogger(__name__)
from vtiles.utils.geopreocessing import flip_y
import vtiles.utils.mercantile as mercantile

def mbtiles_connect(mbtiles_file):
  try:
    conn = sqlite3.connect(mbtiles_file)
    return conn
  except Exception as e:
    logger.error("Could not connect to MBTiles file")
    logger.exception(e)
    sys.exit(1)

def mbtiles_setup(cursor):
  cursor.execute("""
    create table tiles (
      zoom_level integer,
      tile_column integer,
      tile_row integer,
      tile_data blob);
      """)
  cursor.execute("""create table metadata (name text, value text);""")
  cursor.execute("""create unique index name on metadata (name);""")
  cursor.execute("""create unique index tile_index on tiles(zoom_level, tile_column, tile_row);""")

def import_metadata(cursor, metadata_json):
  for name, value in metadata_json.items():
    cursor.execute('insert into metadata (name, value) values (?, ?)',
                (name, value))

def get_zoom_levels(cursor):   
    # Query to get min and max zoom levels
    cursor.execute('''
        SELECT MIN(zoom_level) AS min_zoom, MAX(zoom_level) AS max_zoom
        FROM tiles
    ''')    
    result = cursor.fetchone()    
    min_zoom = result[0] if result else None
    max_zoom = result[1] if result else None  

    return min_zoom, max_zoom

def get_bounds_at_zoom(cursor, zoom_level):
    # Query tiles at the specified zoom level
    cursor.execute("SELECT tile_column, tile_row FROM tiles WHERE zoom_level = ?", (zoom_level,))
    tiles = cursor.fetchall()

    # Calculate bounding boxes for each tile
    bounds = []
    for tile in tiles:
        x, y = tile
        flip_y = (1 << zoom_level) - 1 - y # TMS scheme
        # Calculate bounds for the given tile coordinates
        tile_bounds = mercantile.bounds(x,flip_y, zoom_level)
        bounds.append(tile_bounds)

    return bounds

def compute_max_bound(bounds):
    # Initialize min and max coordinates with extreme values
    min_lat = min_lon = float('inf')
    max_lat = max_lon = float('-inf')

    for bound in bounds:
        # Unpack bounding box coordinates (west, south, east, north)
        west, south, east, north = bound

        # Update min and max values
        min_lon = min(min_lon, west)
        max_lon = max(max_lon, east)    
        min_lat = min(min_lat, south)
        max_lat = max(max_lat, north)
            
    center_lon = (min_lon + max_lon) / 2
    center_lat = (min_lat + max_lat) / 2

    # Return the overall bounding box
    return min_lon, min_lat, max_lon, max_lat,center_lon,center_lat

def get_bounds_center(cursor):   
    boundsString, centerString = None, None
    try:    
        _,max_zoom = get_zoom_levels(cursor)
        bounds_at_max_zoom = get_bounds_at_zoom(cursor, max_zoom)
        bounds = compute_max_bound(bounds_at_max_zoom)
        boundsString = ','.join(map(str, bounds[:4]))
        centerString = ','.join(map(str, bounds[4:]))+ f',{max_zoom}'     
        return boundsString, centerString
    except Exception as e:
        logging.error(f"Get bounds and center erros: {e}")        
        boundsString = '-180.000000,-85.051129,180.000000,85.051129'
        centerString = '0,0,0'
        return boundsString, centerString


# Create metadata table in case no metadata.json found in the tiles folder
def create_metadata(cursor, name, format, tms=0): 
  try:
    minzoom, maxzoom = get_zoom_levels(cursor)
    bounds, center = get_bounds_center(cursor)
    cursor.executemany("INSERT INTO metadata (name, value) VALUES (?, ?);", [
      ("name", name),
      ("format", format), 
      ("bounds", bounds), 
      ("center", center), 
      ("minzoom", minzoom), 
      ("maxzoom", maxzoom), 
      ("desccription", 'MBTiles converted from a tiles folder using vtiles'), 
      ("attribution", '<a href="https://github.com/thangqd/vtiles" target="_blank">&copy; vtiles</a>'),
      ("type", ''),
      ("version", '1')
    ])
    logger.info('Creating metadata done.')
  except Exception as e:
      print(f"Error creating metadata: {e}")

def optimize_connection(cur):
  cur.execute("""PRAGMA synchronous=0""")
  # cur.execute("""PRAGMA locking_mode=EXCLUSIVE""")
  cur.execute("""PRAGMA journal_mode=DELETE""")

def get_dirs(path):
  return [name for name in os.listdir(path)
    if os.path.isdir(os.path.join(path, name))]

def folder2mbtiles(input_folder, mbtiles_file, tms=0):
  logger.info("Converting folder to MBTiles")
  logger.debug("%s --> %s" % (input_folder, mbtiles_file))
  con = mbtiles_connect(mbtiles_file)
  cur = con.cursor()
  optimize_connection(cur)
  mbtiles_setup(cur)
  metadata_json = os.path.join(input_folder, 'metadata.json')
 
  if os.path.exists(metadata_json):
    metadata = json.load(open(metadata_json, 'r'))
    import_metadata(cur, metadata)
    logger.info('Importing metadata done.')
  else:
    logger.warning('metadata.json not found')

  # count = 0
  # start_time = time.time()
  
  with tqdm(desc="Coverting tiles", unit=" tiles") as pbar:
    for zoom_dir in get_dirs(input_folder):   
      z = int(zoom_dir)
      for row_dir in get_dirs(os.path.join(input_folder, zoom_dir)):
        x = int(row_dir)
        for current_file in os.listdir(os.path.join(input_folder, zoom_dir, row_dir)):
          if current_file == ".DS_Store":
            logger.warning("The .DS_Store file will be ignored.")
          else:
            # file_name, ext = current_file.split('.',1)
            file_name, ext = os.path.splitext(current_file)       
            if ext in ('.png','.jpg','.jpeg','.webp', '.pbf'):
              f = open(os.path.join(input_folder, zoom_dir, row_dir, current_file), 'rb')
              file_content = f.read()
              f.close()
              if tms == 1:
                y = flip_y(int(z), int(file_name))
              else:
                y = int(file_name)
              logger.debug(' Read tile from Zoom (z): %i\tCol (x): %i\tRow (y): %i' % (z, x, y))
              cur.execute("""insert into tiles (zoom_level,
                  tile_column, tile_row, tile_data) values
                  (?, ?, ?, ?);""",
                  # (z, x, y, file_content))
                  (z, x, y, sqlite3.Binary(file_content)))
              pbar.update(1)
              # count = count + 1
              # if (count % 100) == 0:
              #   logger.info(" %s tiles inserted (%d tiles/sec)" % (count, count / (time.time() - start_time)))
  try:
    name = os.path.basename(input_folder)
    create_metadata(cur, name, ext, tms)    
  except:      
    logger.warning('No metadata created!')
  
  cur.close()
  con.commit()
  con.close()
    
  logger.info('Converting Folder to MBTiles done.')


def main():
  logging.basicConfig(level=logging.INFO, format='%(message)s')
  parser = argparse.ArgumentParser(description='Convert Tiles folder to MBTiles')
  parser.add_argument('-i', required=True, help='Input folder')
  parser.add_argument('-o', default=None, help='Output mbtiles file name (optional)')
  parser.add_argument('-flipy', type=int, default=0,choices=[0, 1], help='TMS <--> XYZ tiling scheme (optional): 1 or 0, default is 0')

  args = parser.parse_args()

  if not args.i:
    print('Please provide the input folder.')
    exit()

  if not os.path.exists(args.i) or not os.path.isdir(args.i):
    print('Input folder does not exist or invalid!. Please recheck and input a correct one.')
    exit()

  input_folder_abspath =  os.path.abspath(args.i)

  if not args.o:    
    output_file_name = os.path.basename(input_folder_abspath) + '.mbtiles' # Get input folder name
    output_file_abspath  = os.path.abspath(output_file_name)  # Get absolute path of default output file name
    if not os.path.exists(output_file_abspath): 
      print(f'Converting folder {input_folder_abspath} to {output_file_abspath}')        
    else: # the output file is already existed.
      print(f'Output MBTiles file {output_file_abspath} already existed! Please recheck and input a correct one. Ex: -o tiles.mbtiles')
      exit()
  
  else:      
    output_file_abspath = os.path.abspath(args.o)
    if output_file_abspath.endswith('mbtiles') and not os.path.exists(output_file_abspath):
      print(f'Converting folder {input_folder_abspath} to {output_file_abspath}')     
    else:
      print(f'Output MBTiles file {output_file_abspath} is not valid or already existed!. Please recheck and input a correct one. Ex: -o tiles.mbtiles')
      exit()
  
  folder2mbtiles(input_folder_abspath, output_file_abspath, args.flipy)

if __name__ == "__main__":
  main()



import argparse
import json
import os
from .pmtiles.reader import Reader, MmapSource, all_tiles  
from tqdm import tqdm 

def pmtiles_to_folder(input_file, output_folder):
    """
    Convert PMTiles to a folder structure.

    Parameters:
    input_file (str): Path to the input PMTiles file.
    output_folder (str): Path to the output directory.
    """
    os.makedirs(output_folder, exist_ok=True)  # Create output folder if it doesn't exist
    with open(input_file, "r+b") as f:
        source = MmapSource(f)
        reader = Reader(source)
        
        # Write metadata to a JSON file
        with open(os.path.join(output_folder, "metadata.json"), "w") as metadata_file:
            metadata_file.write(json.dumps(reader.metadata()))

        # Get total number of tiles for progress bar
        total_tiles = len(list(all_tiles(source)))
        
        # Iterate over all tiles and write them to the output folder
        for zxy, tile_data in tqdm(all_tiles(source), total=total_tiles, desc="Processing tiles"):
            z, x, y = zxy
            directory = os.path.join(output_folder, str(z), str(x))
            os.makedirs(directory, exist_ok=True)
            path = os.path.join(directory, str(y) + ".pbf")  # Change extension as needed
            with open(path, "wb") as tile_file:
                tile_file.write(tile_data)

def main():
    """
    Main function to parse arguments and call the conversion function.
    """
    parser = argparse.ArgumentParser(description="Convert PMTiles to folder.")
    parser.add_argument("-i", "--input", required=True, help="Input PMTiles file path.")
    parser.add_argument("-o", "--output", required=True, help="Output directory path.")
    args = parser.parse_args()

    input_pmtiles = args.input
    output_directory = args.output

    pmtiles_to_folder(input_pmtiles, output_directory)

if __name__ == "__main__":
    main()



from numbers import Number

from shapely.geometry import shape as shapely_shape
from shapely.geometry.base import BaseGeometry
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry.polygon import Polygon, orient
from shapely.ops import transform
from shapely.wkb import loads as load_wkb
from shapely.wkt import loads as load_wkt

from .geom_encoder import GeometryEncoder
from .Mapbox import vector_tile_pb2 as vector_tile
from .polygon import make_it_valid
from .utils import get_encode_options


def on_invalid_geometry_raise(shape):
    raise ValueError(f"Invalid geometry: {shape.wkt}")


def on_invalid_geometry_ignore(shape):
    return None


def on_invalid_geometry_make_valid(shape):
    return make_it_valid(shape)


class VectorTile:
    def __init__(self, default_options=None):
        self.tile = vector_tile.tile()
        self.default_options = default_options

        self.layer = None
        self.layer_options = None
        self.key_idx = 0
        self.val_idx = 0
        self.seen_keys_idx = {}
        self.seen_values_idx = {}
        self.seen_values_bool_idx = {}
        self.seen_layer_names = set()

    def add_layer(self, name, features, options=None):
        if not name:
            raise ValueError(f"A layer name can not be empty. {name!r} was provided.")
        if name in self.seen_layer_names:
            raise ValueError(f"The layer name {name!r} already exists in the vector tile.")
        self.seen_layer_names.add(name)
        self.layer = self.tile.layers.add()
        self.layer_options = get_encode_options(layer_options=options, default_options=self.default_options)
        self.layer.name = name
        self.layer.version = 2
        self.layer.extent = self.layer_options["extents"]

        self.key_idx = 0
        self.val_idx = 0
        self.seen_keys_idx = {}
        self.seen_values_idx = {}
        self.seen_values_bool_idx = {}

        for feature in features:
            # skip missing or empty geometries
            geometry_spec = feature.get("geometry")
            if geometry_spec is None:
                continue
            shape = self._load_geometry(geometry_spec)

            if shape is None:
                raise NotImplementedError("Can't do geometries that are not wkt, wkb, or shapely geometries")

            if shape.is_empty:
                continue

            if self.layer_options["quantize_bounds"]:
                shape = self.quantize(shape)
            if self.layer_options["check_winding_order"]:
                shape = self.enforce_winding_order(shape)

            if shape is not None and not shape.is_empty:
                self.add_feature(feature, shape)

    def enforce_winding_order(self, shape, n_try=1):
        if shape.geom_type == "MultiPolygon":
            # If we are a multipolygon, we need to ensure that the winding orders of the constituent polygons are
            # correct. In particular, the winding order of the interior rings need to be the opposite of the exterior
            # ones, and all interior rings need to follow the exterior one. This is how the end of one polygon and
            # the beginning of another are signaled.
            shape = self.enforce_multipolygon_winding_order(shape=shape, n_try=n_try)

        elif shape.geom_type == "Polygon":
            # Ensure that polygons are also oriented with the appropriate winding order. Their exterior rings must
            # have a clockwise order, which is translated into a clockwise order in MVT's tile-local coordinates with
            # the Y axis in "screen" (i.e: +ve down) configuration. Note that while the Y axis flips, we also invert
            # the Y coordinate to get the tile-local value, which means the clockwise orientation is unchanged.
            shape = self.enforce_polygon_winding_order(shape=shape, n_try=n_try)

        # other shapes just get passed through
        return shape

    def quantize(self, shape):
        minx, miny, maxx, maxy = self.layer_options["quantize_bounds"]
        extents = self.layer_options["extents"]

        def fn(x, y, z=None):
            xfac = extents / (maxx - minx)
            yfac = extents / (maxy - miny)
            x = xfac * (x - minx)
            y = yfac * (y - miny)
            return round(x), round(y)

        return transform(fn, shape)

    def handle_shape_validity(self, shape, n_try):
        if shape.is_valid:
            return shape

        if n_try >= self.layer_options["max_geometry_validate_tries"]:
            # ensure that we don't recurse indefinitely with an invalid geometry handler that doesn't validate
            # geometries
            return None

        if self.layer_options["on_invalid_geometry"]:
            shape = self.layer_options["on_invalid_geometry"](shape)
            if shape is not None and not shape.is_empty:
                # This means that we have a handler that might have altered the geometry. We'll run through the process
                # again, but keep track of which attempt we are on to terminate the recursion.
                shape = self.enforce_winding_order(shape=shape, n_try=n_try + 1)

        return shape

    def enforce_multipolygon_winding_order(self, shape, n_try):
        assert shape.geom_type == "MultiPolygon"

        parts = []
        for part in shape.geoms:
            part = self.enforce_polygon_winding_order(shape=part, n_try=n_try)
            if part is not None and not part.is_empty:
                if part.geom_type == "MultiPolygon":
                    parts.extend(part.geoms)
                else:
                    parts.append(part)

        if not parts:
            return None

        oriented_shape = parts[0] if len(parts) == 1 else MultiPolygon(parts)
        oriented_shape = self.handle_shape_validity(oriented_shape, n_try)
        return oriented_shape

    def enforce_polygon_winding_order(self, shape, n_try):
        assert shape.geom_type == "Polygon"

        def fn(point):
            x, y = point
            return round(x), round(y)

        exterior = self.apply_map(fn, shape.exterior.coords)
        rings = None

        if len(shape.interiors) > 0:
            rings = [self.apply_map(fn, ring.coords) for ring in shape.interiors]

        sign = 1.0 if self.layer_options["y_coord_down"] else -1.0
        oriented_shape = orient(Polygon(exterior, rings), sign=sign)
        oriented_shape = self.handle_shape_validity(oriented_shape, n_try)
        return oriented_shape

    @staticmethod
    def apply_map(fn, x):
        return list(map(fn, x))

    def _load_geometry(self, geometry_spec):
        if isinstance(geometry_spec, BaseGeometry):
            geom = geometry_spec
        elif isinstance(geometry_spec, dict):
            geom = shapely_shape(geometry_spec)
        else:
            try:
                geom = load_wkb(geometry_spec)
            except Exception:
                try:
                    geom = load_wkt(geometry_spec)
                except Exception:
                    return None

        if self.layer_options["transformer"] is None:
            return geom
        else:
            return transform(self.layer_options["transformer"], geom)

    def add_feature(self, feature, shape):
        geom_encoder = GeometryEncoder(self.layer_options["y_coord_down"], self.layer_options["extents"])
        geometry = geom_encoder.encode(shape)

        feature_type = self._get_feature_type(shape)
        if len(geometry) == 0:
            # Don't add geometry if it's too small
            return
        f = self.layer.features.add()

        fid = feature.get("id")
        if fid is not None and isinstance(fid, Number) and fid >= 0:
            f.id = fid

        # properties
        properties = feature.get("properties")
        if properties is not None:
            self._handle_attr(self.layer, f, properties)

        f.type = feature_type
        f.geometry.extend(geometry)

    def _get_feature_type(self, shape):
        if shape.geom_type == "Point" or shape.geom_type == "MultiPoint":
            return self.tile.Point
        elif shape.geom_type == "LineString" or shape.geom_type == "MultiLineString":
            return self.tile.LineString
        elif shape.geom_type == "Polygon" or shape.geom_type == "MultiPolygon":
            return self.tile.Polygon
        elif shape.geom_type == "GeometryCollection":
            raise ValueError("Encoding geometry collections not supported")
        else:
            raise ValueError(f"Cannot encode unknown geometry type: {shape.geom_type}")

    @staticmethod
    def _chunker(seq, size):
        return [seq[pos : pos + size] for pos in range(0, len(seq), size)]

    @staticmethod
    def _can_handle_key(k):
        return isinstance(k, str)

    @staticmethod
    def _can_handle_val(v):
        return isinstance(v, (str, bool, int, float))

    @classmethod
    def _can_handle_attr(cls, k, v):
        return cls._can_handle_key(k) and cls._can_handle_val(v)

    def _handle_attr(self, layer, feature, props):
        for k, v in props.items():
            if self._can_handle_attr(k, v):
                if k not in self.seen_keys_idx:
                    layer.keys.append(k)
                    self.seen_keys_idx[k] = self.key_idx
                    self.key_idx += 1

                feature.tags.append(self.seen_keys_idx[k])

                values_idx = self.seen_values_bool_idx if isinstance(v, bool) else self.seen_values_idx
                if v not in values_idx:
                    values_idx[v] = self.val_idx
                    self.val_idx += 1

                    val = layer.values.add()
                    if isinstance(v, bool):
                        val.bool_value = v
                    elif isinstance(v, str):
                        val.string_value = v
                    elif isinstance(v, int):
                        val.int_value = v
                    elif isinstance(v, float):
                        val.double_value = v

                feature.tags.append(values_idx[v])



import sqlite3
import vtiles.utils.mercantile as mercantile

def get_max_zoom(mbtiles_input):
    # Connect to the MBTiles SQLite database
    conn = sqlite3.connect(mbtiles_input)
    cursor = conn.cursor()

    # Query to find the minimum zoom level from the tiles table
    cursor.execute("SELECT MAX(zoom_level) FROM tiles")
    min_zoom = cursor.fetchone()[0]

    conn.close()
    return min_zoom

def get_bounds_at_zoom(mbtiles_input, zoom_level):
    # Connect to the MBTiles SQLite database
    conn = sqlite3.connect(mbtiles_input)
    cursor = conn.cursor()

    # Query tiles at the specified zoom level
    cursor.execute("SELECT tile_column, tile_row FROM tiles WHERE zoom_level = ?", (zoom_level,))
    tiles = cursor.fetchall()

    # Calculate bounding boxes for each tile
    bounds = []
    for tile in tiles:
        x, y = tile
        flip_y = (1 << zoom_level) - 1 - y # TMS scheme
        # Calculate bounds for the given tile coordinates
        tile_bounds = mercantile.bounds(x,y, zoom_level)
        bounds.append(tile_bounds)

    conn.close()
    return bounds

def compute_max_bound(bounds):
    # Initialize min and max coordinates with extreme values
    min_lat = min_lon = float('inf')
    max_lat = max_lon = float('-inf')

    for bound in bounds:
        # Unpack bounding box coordinates (west, south, east, north)
        west, south, east, north = bound

        # Update min and max values
        min_lat = min(min_lat, south)
        max_lat = max(max_lat, north)
        min_lon = min(min_lon, west)
        max_lon = max(max_lon, east)

    # Return the overall bounding box
    return min_lon, min_lat, max_lon, max_lat

# Main execution
mbtiles_input = './data/nepal.mbtiles'
min_zoom = get_max_zoom(mbtiles_input)
print(f"Maximum Zoom Level: {min_zoom}")

bounds = get_bounds_at_zoom(mbtiles_input, min_zoom)
max_bound = compute_max_bound(bounds)
print(f"Maximum Bounding Box: {max_bound}")



import math

from geojson2vt.simplify import simplify
from geojson2vt.feature import Slice, create_feature

# converts GeoJSON feature into an intermediate projected JSON vector format with simplification data


def convert(data, options):
    features = []
    if data.get('type') == 'FeatureCollection':
        for i in range(len(data.get('features'))):
            convert_feature(features, data.get('features')[i], options, i)
    elif data.get('type') == 'Feature':
        convert_feature(features, data, options)
    else:
        # single geometry or a geometry collection
        convert_feature(features, {"geometry": data}, options)
    return features


def convert_feature(features, geojson, options, index=None):
    if geojson.get('geometry', None) is None:
        return

    coords = geojson.get('geometry').get('coordinates')
    type_ = geojson.get('geometry').get('type')
    tolerance = math.pow(options.get(
        'tolerance') / ((1 << options.get('maxZoom')) * options.get('extent')), 2)
    geometry = Slice([])
    id_ = geojson.get('id')
    if options.get('promoteId', None) is not None and geojson.get('properties', None) is not None and 'promoteId' in geojson.get('properties'):
        id_ = geojson['properties'][options.get('promoteId')]
    elif options.get('generateId', False):
        id_ = index if index is not None else 0

    if type_ == 'Point':
        convert_point(coords, geometry)
    elif type_ == 'MultiPoint':
        for p in coords:
            convert_point(p, geometry)
    elif type_ == 'LineString':
        convert_line(coords, geometry, tolerance, False)
    elif type_ == 'MultiLineString':
        if options.get('lineMetrics'):
            # explode into linestrings to be able to track metrics
            for line in coords:
                geometry = Slice([])
                convert_line(line, geometry, tolerance, False)
                features.append(create_feature(id_, 'LineString',
                                               geometry, geojson.get('properties')))
            return
        else:
            convert_lines(coords, geometry, tolerance, False)
    elif type_ == 'Polygon':
        convert_lines(coords, geometry, tolerance, True)
    elif type_ == 'MultiPolygon':
        for polygon in coords:
            newPolygon = []
            convert_lines(polygon, newPolygon, tolerance, True)
            geometry.append(newPolygon)
    elif type_ == 'GeometryCollection':
        for singleGeometry in geojson['geometry']['geometries']:
            convert_feature(features, {
                "id": str(id_),
                "geometry": singleGeometry,
                "properties": geojson.get('properties')
            }, options, index)
        return
    else:
        raise Exception('Input data is not a valid GeoJSON object.')

    features.append(create_feature(
        id_, type_, geometry, geojson.get('properties')))


def convert_point(coords, out):
    out.append(project_x(coords[0]))
    out.append(project_y(coords[1]))
    out.append(0)


def convert_line(ring, out, tolerance, isPolygon):
    x0, y0 = None, None
    size = 0

    for j in range(len(ring)):
        x = project_x(ring[j][0])
        y = project_y(ring[j][1])

        out.append(x)
        out.append(y)
        out.append(0)

        if j > 0:
            if isPolygon:
                size += (x0 * y - x * y0) / 2  # area
            else:
                size += math.sqrt(math.pow(x - x0, 2) +
                                  math.pow(y - y0, 2))  # length
        x0 = x
        y0 = y

    last = len(out) - 3
    out[2] = 1
    simplify(out, 0, last, tolerance)
    out[last + 2] = 1.

    out.size = abs(size)
    out.start = 0.
    out.end = out.size


def convert_lines(rings, out, tolerance, isPolygon):
    for i in range(len(rings)):
        geom = Slice([])
        convert_line(rings[i], geom, tolerance, isPolygon)
        out.append(geom)


def project_x(x):
    return x / 360. + 0.5


def project_y(y):
    sin = math.sin(y * math.pi / 180.)
    if sin == 1.:
        return 0.
    if sin == -1.:
        return 1.
    y2 = 0.5 - 0.25 * math.log((1. + sin) / (1. - sin)) / math.pi
    return 0 if y2 < 0. else (1. if y2 > 1. else y2)



import argparse, os
import sqlite3
import gzip, zlib
import logging
import shutil
from tqdm import tqdm
from vtiles.mbtiles import mbtilesfixmeta, mbtilesdelduplicate

def compress_tile_data(tile_data):
    try: 
        if tile_data[:2] != b'\x1f\x8b':  # check if not GZIP    
            if tile_data[:2] == b'\x78\x9c' or tile_data[:2] == b'\x78\x01' or tile_data[:2] == b'\x78\xda': # Check if ZLIB
                tile_data = zlib.decompress(tile_data)           
            tile_data = gzip.compress(tile_data)
    except Exception as e:
        logging.error(f"Failed to compress tile data: {e}")
        return tile_data
    return tile_data          

def compress_mbtiles(input_mbtiles, output_mbtiles):
    # if os.path.exists(output_mbtiles):
    #     os.remove(output_mbtiles)
    # shutil.copyfile(input_mbtiles, output_mbtiles)
    mbtilesdelduplicate.remove_duplicates(input_mbtiles,output_mbtiles)
    # Open the copied MBTiles file
    conn = sqlite3.connect(output_mbtiles)
    cursor = conn.cursor()

    # Check if the tiles table is a view
    cursor.execute("SELECT type FROM sqlite_master WHERE name='tiles'")
    result = cursor.fetchone()

    if result and result[0] == 'view':
        # Create a new table named tiles_new
        cursor.execute("CREATE TABLE tiles_new (zoom_level INTEGER, tile_column INTEGER, tile_row INTEGER, tile_data BLOB)")
        
        # Select data from the view
        cursor.execute("SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles order by zoom_level")
        tiles = cursor.fetchall()
        
        # Insert decompressed data into the new table
        for zoom_level, tile_column, tile_row, tile_data in tqdm(tiles, desc="Compressing tiles", unit="tile"):
            try:
                compressed_tile = compress_tile_data(tile_data)
                cursor.execute(
                    "INSERT INTO tiles_new (zoom_level, tile_column, tile_row, tile_data) VALUES (?, ?, ?, ?)",
                    (zoom_level, tile_column, tile_row,compressed_tile)
                )
            except Exception as e:
                print(f"Error compressing tile {zoom_level}/{tile_column}/{tile_row}: {e}")
    
        # Drop the view and rename the new table to tiles
        cursor.execute("DROP VIEW tiles")
        cursor.execute("ALTER TABLE tiles_new RENAME TO tiles")
        cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS tile_index ON tiles (zoom_level, tile_column, tile_row)")
    else:
        # Decompress tile data in the existing tiles table
        cursor.execute("SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles order by zoom_level")
        tiles = cursor.fetchall()
        
        # Add tqdm progress bar
        for zoom_level, tile_column, tile_row, tile_data in tqdm(tiles, desc="Compressing tiles", unit="tile"):
            compressed_tile = compress_tile_data(tile_data)
            try:
                cursor.execute(
                    "UPDATE tiles SET tile_data = ? WHERE zoom_level = ? AND tile_column = ? AND tile_row = ?",
                    (compressed_tile, zoom_level, tile_column, tile_row)
                )
            except Exception as e:
                print(f"Error compressing tile {zoom_level}/{tile_column}/{tile_row}: {e}")
    
        cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS tile_index ON tiles (zoom_level, tile_column, tile_row)")
    # Commit and close connections
    conn.commit()
    conn.close()
   
def main():
    parser = argparse.ArgumentParser(description='Compress MBTiles file with GZIP.')
    parser.add_argument('-i', '--input', required=True, help='Path to the input MBTiles file.')
    parser.add_argument('-o', '--output', required=True, help='Path to the output MBTiles file.')

    args = parser.parse_args()
    compress_mbtiles(args.input, args.output)

if __name__ == "__main__":
    main()



import os, requests, enum
from tqdm import tqdm
import urllib
import tempfile
from urllib.error import HTTPError
from urllib.request import *
import argparse

from .geofabrik_sources import (
        Africa,
        Antarctica,
        Asia,
        AustraliaOceania,
        Europe,
        NorthAmerica,
        SouthAmerica,
        CentralAmerica,
        Brazil,
        Canada,
        France,
        Germany,
        GreatBritain,
        Italy,
        Japan,
        Netherlands,
        Poland,
        Russia,
        USA,
        SubRegions,
    )

class DataSources:
    def __init__(self):
        self.africa = Africa()
        self.antarctica = Antarctica()
        self.asia = Asia()
        self.australia_oceania = AustraliaOceania()
        self.europe = Europe()
        self.north_america = NorthAmerica()
        self.south_america = SouthAmerica()
        self.central_america = CentralAmerica()
        self.subregions = SubRegions()

        self.available = {
            "africa": self.africa.available,
            "antarctica": self.antarctica.available,
            "asia": self.asia.available,
            "australia_oceania": self.australia_oceania.available,
            "central_america": self.central_america.available,
            "europe": self.europe.available,
            "north_america": self.north_america.available,
            "south_america": self.south_america.available,
            "subregions": self.subregions.available,
        }

        # Gather all data sources
        # Keep hidden to avoid encouraging iteration of the whole
        # world at once which most likely would end up
        # in memory error / filling the disk etc.
        self._all_sources = [
            k for k in self.available.keys() if k not in ["subregions"]
        ]

        for source, available in self.available.items():
            self._all_sources += available

        for subregion in self.subregions.available:
            self._all_sources += self.subregions.__dict__[subregion].available

        self._all_sources = [src.lower() for src in self._all_sources]      
        self._all_sources = list(set(self._all_sources))

class UNIT(enum.Enum):
    BYTES = 1
    KB = 2
    MB = 3
    GB = 4

def convert_unit(size_in_bytes, unit):
    if unit == UNIT.KB:
        return size_in_bytes / 1024
    elif unit == UNIT.MB:
        return size_in_bytes / (1024 * 1024)
    elif unit == UNIT.GB:
        return size_in_bytes / (1024 * 1024 * 1024)
    else:
        return size_in_bytes


def get_file_size(file_name, size_type=UNIT.MB):
    size = os.path.getsize(file_name)
    return round(convert_unit(size, size_type), 2)


def download(url, filename, update, target_dir):
    if target_dir is None:
        temp_dir = tempfile.gettempdir()
        target_dir = os.path.join(temp_dir, "vtiles")
    else:
        if not os.path.isdir(target_dir):
            raise ValueError(f"The provided directory does not exist: " f"{target_dir}")

    filepath = os.path.abspath(os.path.join(target_dir, os.path.basename(filename)))

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    # Check if file exists
    file_exists = False
    if os.path.exists(filepath):
        file_exists = True

    if update and file_exists:
        os.remove(filepath)

    # Download data to temp if it does not exist or if update is requested
    if update or file_exists is False:
        try:
            filepath, msg = urllib.request.urlretrieve(url, filepath)
        except HTTPError:
            raise ValueError(
                f"PBF-file '{url}' is temporarily unavailable. " f"Try again later."
            )
        except Exception as e:
            raise e

        filesize = get_file_size(filepath)
        if filesize == 0:
            raise ValueError(
                f"PBF-file '{filename}' from the provider was empty. "
                "This is likely a temporary issue, try again later."
            )
        print(
            f"Downloaded Protobuf data '{os.path.basename(filepath)}' "
            f"({filesize} MB) to:\n'{filepath}'"
        )
    return filepath

# Initialize DataSources
sources = DataSources()
available = {
    "regions": {
        k: v for k, v in sources.available.items() if k not in ["subregions"]
    },
    "subregions": sources.subregions.available
}


def retrieve(data, update, directory):
    # return download(
    #     url=data["url"], filename=data["name"], update=update, target_dir=directory
    # )
    return download_file(data["url"], os.getcwd())


def search_source(name):
    for source, available in sources.available.items():
        if isinstance(available, list):
            if name in available:
                return sources.__dict__[source].__dict__[name]
        elif isinstance(available, dict):
            # Sub-regions should be looked one level further down
            for subregion, available2 in available.items():
                if name in available2:
                    return sources.subregions.__dict__[subregion].__dict__[name]
    raise ValueError(f"Could not retrieve url for '{name}'.")

def print_available_datasets():
    for category, datasets in available.items():
        print(f"\nCategory: '{category}':")
        if isinstance(datasets, dict):
            for subcategory, subdatasets in datasets.items():
                print(f"subCategory: '{subcategory}':\n")
                print(", ".join(subdatasets))
                input("\nPress Enter to continue...\n")
        else:
            print(", ".join(datasets))
        

# def get_data(dataset, update=False, directory=None):
#     """
#     Get the path to a PBF data file, and download the data if needed.

#     Parameters
#     ----------
#     dataset : str
#         The name of the dataset. Run ``pyrosm.data.available`` for
#         all available options.

#     update : bool
#         Whether the PBF file should be downloaded/updated if the dataset
#         with the same name exists in the temp.

#     directory : str (optional)
#         Path to a directory where the PBF data will be downloaded.
#         (does not apply for test data sets bundled with the package).
#     """
#     if not isinstance(dataset, str):
#         raise ValueError(f"'dataset' should be text. Got {dataset}.")
#     dataset = dataset.lower().strip()

#     if dataset in sources._all_sources:
#         return retrieve(search_source(dataset), update, directory)

#     elif dataset.replace(" ", "") in sources._all_sources:
#         return retrieve(search_source(dataset.replace(" ", "")), update, directory)

#     # Users might pass country names without underscores (e.g. North America)
#     elif dataset.replace(" ", "_") in sources._all_sources:
#         return retrieve(search_source(dataset.replace(" ", "_")), update, directory)

#     # Users might pass country names with dashes instead of underscores (e.g. canary-islands)
#     elif dataset.replace("-", "_") in sources._all_sources:
#         return retrieve(search_source(dataset.replace("-", "_")), update, directory)

#     else:
#         msg = "The dataset '{data}' is not available. ".format(data=dataset)
#         msg += "Available datasets are: {}".format(", ".join(sources._all_sources))
#         raise ValueError(msg)

def get_data(dataset, update=False, directory=None):
    """
    Get the path to a PBF data file, and download the data if needed.

    Parameters
    ----------
    dataset : str
        The name of the dataset. Run ``print_available_datasets`` for
        all available options.

    update : bool
        Whether the PBF file should be downloaded/updated if the dataset
        with the same name exists in the temp.

    directory : str (optional)
        Path to a directory where the PBF data will be downloaded.
        (does not apply for test data sets bundled with the package).
    """
    if not isinstance(dataset, str):
        raise ValueError(f"'dataset' should be text. Got {dataset}.")
    dataset = dataset.lower().strip()

    if dataset in sources._all_sources:
        return retrieve(search_source(dataset), update, directory)

    elif dataset.replace(" ", "") in sources._all_sources:
        return retrieve(search_source(dataset.replace(" ", "")), update, directory)

    elif dataset.replace(" ", "_") in sources._all_sources:
        return retrieve(search_source(dataset.replace(" ", "_")), update, directory)

    elif dataset.replace("-", "_") in sources._all_sources:
        return retrieve(search_source(dataset.replace("-", "_")), update, directory)

    else:
        msg = "The dataset '{}' is not available. ".format(dataset)
        msg+= "Available datasets are: "
        print(msg)
        # input("(Press Enter to continue...)")
        # msg += "Available datasets are: {}".format(", ".join(sources._all_sources))    
        # raise ValueError(msg)    
        print_available_datasets()        

####### Test
def download_file(url, save_to) :
    # url     - url of downloadable file
    # save_to - directory to save the file to
    
    print("Downloading from:", url)       
    req = requests.get(url, stream=True)
    if 'Content-Disposition' in req.headers:
        filename = req.headers['Content-Disposition'].split("filename=")[1]
    else:
        filename = os.path.basename(url)
    os.chdir(save_to)

    if req.status_code == 200:
        # Calculate the total file size in bytes
        total_size = int(req.headers.get('content-length', 0))
        # Open the file in binary write mode
        with open(filename, 'wb') as file:
            # Use tqdm to show the progress bar
            with tqdm(total=total_size, unit='B', unit_scale=True, desc=filename, ncols=100) as pbar:
                # Iterate over the content of the response in chunks
                for chunk in req.iter_content(chunk_size=8192):
                    if chunk:
                        # Write the chunk to the file
                        file.write(chunk)
                        # Update the progress bar with the size of the chunk
                        pbar.update(len(chunk))
        file.close()
        print(filename, "has been downloaded to", save_to)
    #______________________________________________________________________________

def main():    
    parser = argparse.ArgumentParser(description="Download OSM data by country/ region.")
    parser.add_argument("dataset", type=str, help="The name of the countrty/ region to download.")   
    args = parser.parse_args()
    dataset = args.dataset
    cur_dir = os.getcwd()
    try:
        get_data(dataset, update=True, directory=cur_dir)
    except ValueError as e:
        print(e)

    # save_to = os.getcwd()
    # get_data("yemen", update=True,directory = save_to)
    # print (available)
    # print (available.keys())
    # url = "https://github.com/udsleeds/openinfra/releases/download/v0.2/Leeds_06_06_22.osm.pbf"
    # # Change this to your desired download directory.
    # save_to = os.getcwd()
    # download_file(url, save_to)

if __name__ == "__main__":
    main()




import itertools as it

from .utils import CMD_BITS, CMD_FAKE, CMD_LINE_TO, CMD_MOVE_TO, CMD_SEG_END, zig_zag_encode


class GeometryEncoder:
    def __init__(self, y_coord_down, extents):
        self._geometry = []
        self._y_coord_down = y_coord_down
        self._extents = extents
        self._last_x, self._last_y = 0, 0

    @classmethod
    def encode_cmd_length(cls, cmd, length):
        return (length << CMD_BITS) | (cmd & ((1 << CMD_BITS) - 1))

    @staticmethod
    def omit_last(iterator):
        iterator, lookahead = it.tee(iterator)
        next(lookahead, None)
        for _ in lookahead:
            yield next(iterator)

    def coords_on_grid(self, x, y):
        """Snap coordinates on the grid with integer coordinates"""
        if isinstance(x, float):
            x = int(round(x))
        if isinstance(y, float):
            y = int(round(y))
        if not self._y_coord_down:
            y = self._extents - y
        return x, y

    def encode_multipoint(self, points):
        cmd_move_to = self.encode_cmd_length(CMD_MOVE_TO, len(points))
        self._geometry = [cmd_move_to]
        last_x = 0
        last_y = 0
        for point in points:
            x, y = self.coords_on_grid(point.x, point.y)
            dx, dy = x - last_x, y - last_y
            self._geometry.append(zig_zag_encode(dx))
            self._geometry.append(zig_zag_encode(dy))
            last_x = x
            last_y = y

    def encode_arc(self, coords):
        """Appends commands to _geometry to create an arc.
        - Returns False if nothing was added
        - Returns True and moves _last_x, _last_y if
            some points where added
        """
        last_x, last_y = self._last_x, self._last_y
        float_x, float_y = next(coords)
        x, y = self.coords_on_grid(float_x, float_y)
        dx, dy = x - last_x, y - last_y
        cmd_encoded = self.encode_cmd_length(CMD_MOVE_TO, 1)
        commands = [cmd_encoded, zig_zag_encode(dx), zig_zag_encode(dy), CMD_FAKE]
        pairs_added = 0
        last_x, last_y = x, y
        for float_x, float_y in coords:
            x, y = self.coords_on_grid(float_x, float_y)
            dx, dy = x - last_x, y - last_y
            if dx == 0 and dy == 0:
                continue
            commands.append(zig_zag_encode(dx))
            commands.append(zig_zag_encode(dy))
            last_x, last_y = x, y
            pairs_added += 1
        if pairs_added == 0:
            return False
        cmd_encoded = self.encode_cmd_length(CMD_LINE_TO, pairs_added)
        commands[3] = cmd_encoded
        self._geometry.extend(commands)
        self._last_x, self._last_y = last_x, last_y
        return True

    def encode_multilinestring(self, shape):
        for arc in shape.geoms:
            coords = iter(arc.coords)
            self.encode_arc(coords)

    def encode_ring(self, ring):
        coords = self.omit_last(iter(ring.coords))
        if not self.encode_arc(coords):
            return False
        cmd_seg_end = self.encode_cmd_length(CMD_SEG_END, 1)
        self._geometry.append(cmd_seg_end)
        return True

    def encode_polygon(self, shape):
        if not self.encode_ring(shape.exterior):
            return
        for arc in shape.interiors:
            self.encode_ring(arc)

    def encode_multipolygon(self, shape):
        for polygon in shape.geoms:
            self.encode_polygon(polygon)

    def encode(self, shape):
        if shape.geom_type == "GeometryCollection":
            # do nothing
            pass
        elif shape.geom_type == "Point":
            x, y = self.coords_on_grid(shape.x, shape.y)
            cmd_encoded = self.encode_cmd_length(CMD_MOVE_TO, 1)
            self._geometry = [cmd_encoded, zig_zag_encode(x), zig_zag_encode(y)]
        elif shape.geom_type == "MultiPoint":
            self.encode_multipoint(shape.geoms)
        elif shape.geom_type == "LineString":
            coords = iter(shape.coords)
            self.encode_arc(coords)
        elif shape.geom_type == "MultiLineString":
            self.encode_multilinestring(shape)
        elif shape.geom_type == "Polygon":
            self.encode_polygon(shape)
        elif shape.geom_type == "MultiPolygon":
            self.encode_multipolygon(shape)
        else:
            raise NotImplementedError(f"Can't do {shape.geom_type} geometries")
        return self._geometry



import sqlite3
import argparse
from io import BytesIO
from PIL import Image
from vtiles.utils.mapbox_vector_tile import decode
import gzip
import zlib


def get_sample_tile_data(db_path):
    """Retrieve a sample tile data from the tiles table."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = "SELECT tile_data FROM tiles LIMIT 1;"
    cursor.execute(query)
    tile_data = cursor.fetchone()

    conn.close()
    if tile_data:
        return tile_data[0]
    else:
        return None

def decompress_tile_data(tile_data):
    """Decompress tile data if it is compressed and return decompressed data with a compression flag."""
    try:
        # Check for gzip compression
        with BytesIO(tile_data) as buf:
            buf.seek(0)
            with gzip.GzipFile(fileobj=buf) as gz:
                return gz.read(), 'gzip'
    except (OSError, EOFError):
        try:
            # Check for zlib compression
            return zlib.decompress(tile_data), 'zlib'
        except zlib.error:
            # No compression or unrecognized compression format
            return tile_data, 'none'
    except Exception as e:
        raise RuntimeError(f"Decompression error: {e}")

def check_tile_format(tile_data):
    """Check the format of the tile data and print whether it is compressed."""
    if tile_data:
        try:
            decompressed_data, compression_type = decompress_tile_data(tile_data)  
            print(f"Compression type: {compression_type}")                  
            with BytesIO(decompressed_data) as buf:
                try:
                    # Attempt to open as an image (for raster tiles)
                    with Image.open(buf) as img:
                        return f"Raster Tile (Format: {img.format})"
                except (IOError, OSError):
                    try:
                        # Attempt to decode as vector tile (for vector tiles)
                        decode(decompressed_data)
                        return "Vector Tile"
                    except Exception as e:
                        return f"Error identifying format: {e}"
        except Exception as e:
            return f"Error during decompression or format check: {e}"
    else:
        return "No tile data found."


def get_standard_tile_count(zoom_level):
    """Calculate the standard number of tiles for a given zoom level."""
    num_tiles_per_side = 2 ** zoom_level
    return num_tiles_per_side * num_tiles_per_side

def count_tiles_for_each_zoom(db_path):
    """Count the number of tiles for each zoom level in the tiles table."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = """
    SELECT zoom_level, COUNT(*) AS tile_count
    FROM tiles
    GROUP BY zoom_level
    ORDER BY zoom_level;
    """

    cursor.execute(query)
    results = cursor.fetchall()

    conn.close()
    return results

def find_duplicates(db_path):
    """Find duplicate rows in the tiles table based on zoom_level, tile_column, and tile_row."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = """
    SELECT zoom_level, tile_column, tile_row, COUNT(*) AS count
    FROM tiles
    GROUP BY zoom_level, tile_column, tile_row
    HAVING COUNT(*) > 1
    ORDER BY count;
    """

    cursor.execute(query)
    duplicates = cursor.fetchall()

    conn.close()
    return duplicates

def count_total_duplicates(duplicates):
    """Calculate the total number of duplicate rows."""
    total_duplicates = sum(count - 1 for _, _, _, count in duplicates)
    return total_duplicates

def get_min_zoom_level(db_path):
    """Get the minimum zoom level from the tiles table."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = "SELECT MIN(zoom_level) FROM tiles;"
    cursor.execute(query)
    min_zoom = cursor.fetchone()[0]

    conn.close()
    return min_zoom

def get_max_zoom_level(db_path):
    """Get the maximum zoom level from the tiles table."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = "SELECT MAX(zoom_level) FROM tiles;"
    cursor.execute(query)
    max_zoom = cursor.fetchone()[0]

    conn.close()
    return max_zoom

def count_total_tiles(db_path):
    """Calculate the total number of tiles in the tiles table."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = "SELECT COUNT(*) FROM tiles;"
    cursor.execute(query)
    total_tiles = cursor.fetchone()[0]

    conn.close()
    return total_tiles

def main():
    parser = argparse.ArgumentParser(description='Analyze an MBTiles file for tile counts and duplicates.')
    parser.add_argument('db_path', type=str, help='Path to the MBTiles database file.')

    args = parser.parse_args()

    db_path = args.db_path

    # Retrieve a sample tile data
    tile_data = get_sample_tile_data(db_path)
    
    # Get the tile counts for each zoom level
    tile_counts = count_tiles_for_each_zoom(db_path)
    
    # Get min and max zoom levels
    min_zoom_level = get_min_zoom_level(db_path)
    max_zoom_level = get_max_zoom_level(db_path)
    
    # Get total number of tiles
    total_tiles = count_total_tiles(db_path)
    
    print(f"Min Zoom Level: {min_zoom_level}")
    print(f"Max Zoom Level: {max_zoom_level}")
    print(f"Total Number of Tiles: {total_tiles}")
    # Check the format of the tile data
    tile_format = check_tile_format(tile_data)    
    print(f"Tile format: {tile_format}")

    print("\nTile count for each zoom level:")
    # Print results with standard number of tiles
    print(f"{'Zoom Level':<12} {'Actual Tile Count':<20} {'Standard Tile Count':<20} {'Matches Standard'}")
    print("="*62)

    for zoom_level, actual_tile_count in tile_counts:
        standard_tile_count = get_standard_tile_count(zoom_level)
        matches_standard = "Yes" if actual_tile_count == standard_tile_count else "No"
        print(f"{zoom_level:<12} {actual_tile_count:<20} {standard_tile_count:<20} {matches_standard}")

    # Find duplicates
    duplicates = find_duplicates(db_path)
    
    # Calculate and print total number of duplicate rows
    total_duplicates = count_total_duplicates(duplicates)
    
    # Print duplicates
    print("\nDuplicate Rows:")
    print(f"Total number of duplicate rows: {total_duplicates}")
    if duplicates:
        print(f"{'Zoom Level':<12} {'Tile Column':<12} {'Tile Row':<12} {'Count':<6}")
        print("="*42)
        for zoom_level, tile_column, tile_row, count in duplicates:
            print(f"{zoom_level:<12} {tile_column:<12} {tile_row:<12} {count:<6}")
    else:
        print("No duplicate rows found.")

if __name__ == '__main__':
    main()


import math
from geojson2vt.feature import create_feature, Slice


r""" 
clip features between two vertical or horizontal axis-parallel lines:
 *     |        |
 *  ___|___     |     /
 * /   |   \____|____/
 *     |        |
 *
 * k1 and k2 are the line coordinates
 * axis: 0 for x, 1 for y
 * minAll and maxAll: minimum and maximum coordinate value for all features
 """


def clip(features, scale, k1, k2, axis, minAll, maxAll, options):
    k1 /= scale
    k2 /= scale

    if minAll >= k1 and maxAll < k2:
        return features  # trivial accept
    elif maxAll < k1 or minAll >= k2:
        return None  # trivial reject

    clipped = []

    for feature in features:
        if isinstance(feature.get('geometry'), Slice):
            geometry = feature.get('geometry')
        else:
            geometry = Slice(feature.get('geometry'))

        type_ = feature.get('type')

        min_ = feature.get('minX') if axis == 0 else feature.get('minY')
        max_ = feature.get('maxX') if axis == 0 else feature.get('maxY')

        if min_ >= k1 and max_ < k2:  # trivial accept
            clipped.append(feature)
            continue
        elif max_ < k1 or min_ >= k2:  # trivial reject
            continue

        newGeometry = Slice([])  # []

        if type_ == 'Point' or type_ == 'MultiPoint':
            clip_points(geometry, newGeometry, k1, k2, axis)
        elif type_ == 'LineString':
            clip_line(geometry, newGeometry, k1, k2,
                      axis, False, options.get('lineMetrics', False))
        elif type_ == 'MultiLineString':
            clip_lines(geometry, newGeometry, k1, k2, axis, False)
        elif type_ == 'Polygon':
            if any(isinstance(l, list) for l in geometry):
                clip_lines(geometry, newGeometry, k1, k2, axis, True)
            else:
                clip_line(geometry, newGeometry, k1, k2, axis, True, False)
        elif type_ == 'MultiPolygon':
            for polygon in geometry:
                newPolygon = Slice([])
                clip_lines(polygon, newPolygon, k1, k2, axis, True)
                if len(newPolygon) > 0:
                    newGeometry.append(newPolygon)

        if len(newGeometry) > 0:
            if options.get('lineMetrics', False) and type_ == 'LineString':
                for line in newGeometry:
                    clipped.append(create_feature(
                        feature.get('id'), type_, line, feature.get('tags')))
                continue
            if type_ == 'LineString' or type_ == 'MultiLineString':
                if len(newGeometry) == 1:
                    type_ = 'LineString'
                    newGeometry = newGeometry[0]
                else:
                    type_ = 'MultiLineString'

            if type_ == 'Point' or type_ == 'MultiPoint':
                type_ = 'Point' if len(newGeometry) == 3 else 'MultiPoint'

            clipped.append(create_feature(
                feature.get('id'), type_, newGeometry, feature.get('tags')))

    return clipped if len(clipped) > 0 else None


def clip_points(geom, newGeom, k1, k2, axis):
    for i in range(0, len(geom), 3):
        a = geom[i + axis]
        if a >= k1 and a <= k2:
            add_point(newGeom, geom[i], geom[i + 1], geom[i + 2])


def clip_line(geom, newGeom, k1, k2, axis, isPolygon, trackMetrics):
    slice_ = new_slice(geom)
    intersect = intersectX if axis == 0 else intersectY
    l = geom.start if isinstance(geom, Slice) else 0.
    segLen, t = None, None

    # length = len(geom) if isinstance(geom, list) else 0
    for i in range(0, len(geom) - 3, 3):
        ax = geom[i]
        ay = geom[i + 1]
        az = geom[i + 2]
        bx = geom[i + 3]
        by = geom[i + 4]
        a = ax if axis == 0 else ay
        b = bx if axis == 0 else by
        exited = False

        if trackMetrics:
            segLen = math.sqrt(math.pow(ax - bx, 2) + math.pow(ay - by, 2))

        if a < k1:
            # ---|-->  | (line enters the clip region from the left)
            if b > k1:
                t = intersect(slice_, ax, ay, bx, by, k1)
                if trackMetrics:
                    slice_.start = l + segLen * t
        elif a > k2:
            # |  <--|--- (line enters the clip region from the right)
            if b < k2:
                t = intersect(slice_, ax, ay, bx, by, k2)
                if trackMetrics:
                    slice_.start = l + segLen * t
        else:
            add_point(slice_, ax, ay, az)
        if b < k1 and a >= k1:
            # <--|---  | or <--|-----|--- (line exits the clip region on the left)
            t = intersect(slice_, ax, ay, bx, by, k1)
            exited = True
        if b > k2 and a <= k2:
            # |  ---|--> or ---|-----|--> (line exits the clip region on the right)
            t = intersect(slice_, ax, ay, bx, by, k2)
            exited = True

        if not isPolygon and exited:
            if trackMetrics:
                slice_.end = l + segLen * t
            newGeom.append(slice_)
            slice_ = new_slice(geom)

        if trackMetrics:
            l += segLen

    # add the last point
    last = len(geom) - 3
    ax = geom[last]
    ay = geom[last + 1]
    az = geom[last + 2]
    a = ax if axis == 0 else ay
    if a >= k1 and a <= k2:
        add_point(slice_, ax, ay, az)

    # close the polygon if its endpoints are not the same after clipping
    last = len(slice_) - 3
    if isPolygon and last >= 3 and (slice_[last] != slice_[0] or slice_[last + 1] != slice_[1]):
        add_point(slice_, slice_[0], slice_[1], slice_[2])

    # add the final slice
    if len(slice_) > 0:
        newGeom.append(slice_)


def new_slice(line):
    slice_ = Slice([])
    slice_.size = line.size if isinstance(line, Slice) else 0.
    slice_.start = line.start if isinstance(line, Slice) else 0.
    slice_.end = line.end if isinstance(line, Slice) else 0.
    return slice_


def clip_lines(geom, newGeom, k1, k2, axis, isPolygon):
    for line in geom:
        clip_line(line, newGeom, k1, k2, axis, isPolygon, False)


def add_point(out, x, y, z):
    out.append(x)
    out.append(y)
    out.append(z)


def intersectX(out, ax, ay, bx, by, x):
    t = (x - ax) / (bx - ax)
    add_point(out, x, ay + (by - ay) * t, 1)
    return t


def intersectY(out, ax, ay, bx, by, y):
    t = (y - ay) / (by - ay)
    add_point(out, ax + (bx - ax) * t, y, 1)
    return t



#!/usr/bin/env python
import sys
import pprint
from .pmtiles.reader import Reader, MmapSource

def print_usage():
    print("Usage: pmtilesinfo PMTILES_FILE")
    print("Usage: pmtilesinfo PMTILES_FILE Z X Y")
    exit(1)

def main():
    if len(sys.argv) <= 1:
        print_usage()

    pmtiles_file = sys.argv[1]

    try:
        with open(pmtiles_file, "r+b") as f:
            reader = Reader(MmapSource(f))
            if len(sys.argv) == 2:
                pprint.pprint(reader.header())
                pprint.pprint(reader.metadata())
            elif len(sys.argv) == 5:
                z = int(sys.argv[2])
                x = int(sys.argv[3])
                y = int(sys.argv[4])
                tile_data = reader.get(z, x, y)
                sys.stdout.buffer.write(tile_data)
            else:
                print_usage()
    except Exception as e:
        print(f"Error: {e}")
        exit(1)

if __name__ == "__main__":
    main()



import sqlite3
import os
import shutil
from vtiles.utils.mapbox_vector_tile import encode, decode
from vtiles.utils.geopreocessing import fix_wkt
import argparse
import gzip, zlib
import json
import logging
from tqdm import tqdm
from vtiles.mbtiles import mbtilesfixmeta

logging.basicConfig(level=logging.INFO)

def merge_json_layers(layer1, layer2):
    # Create a dictionary to combine features by layer name
    combined_layer = {}
    
    # Add features from the first JSON data
    for layer in layer1:
        name = layer['name']
        if name not in combined_layer:
            combined_layer[name] = layer
        else:
            combined_layer[name]['features'].extend(layer['features'])

    # Add features from the second JSON data
    for layer in layer2:
        name = layer['name']
        if name not in combined_layer:
            combined_layer[name] = layer
        else:
            combined_layer[name]['features'].extend(layer['features'])

    # Convert combined data to a list
    merged_layer= list(combined_layer.values())
    
    return merged_layer

def merge_tiles(tile1, tile2,z=None, x=None, y=None):
    try:        
        if tile1 and tile2:
            if tile1[:2] == b'\x1f\x8b':
                tile1 = gzip.decompress(tile1)
            elif tile1[:2] == b'\x78\x9c' or tile1[:2] == b'\x78\x01' or tile1[:2] == b'\x78\xda':
                tile1 = zlib.decompress(tile1)
            decoded_tile1 = decode(tile1)
            decoded_tile1_fixed = fix_wkt(decoded_tile1)

            if tile2[:2] == b'\x1f\x8b':
                tile2 = gzip.decompress(tile2)
            elif tile2[:2] == b'\x78\x9c' or tile2[:2] == b'\x78\x01' or tile2[:2] == b'\x78\xda':
                tile2 = zlib.decompress(tile2)
            decoded_tile2 = decode(tile2)
            decoded_tile2_fixed = fix_wkt(decoded_tile2)

            merged_layer = merge_json_layers(decoded_tile1_fixed,decoded_tile2_fixed)
            # print (merged_tiles)
            merged_tile_encoded = encode(merged_layer)
            merged_tile_encoded_gzip = gzip.compress(merged_tile_encoded)   
            return merged_tile_encoded_gzip
                
        elif tile1:
            if tile1[:2] == b'\x1f\x8b':
                return tile1
            elif tile1[:2] == b'\x78\x9c' or tile1[:2] == b'\x78\x01' or tile1[:2] == b'\x78\xda':
                tile1_decompressed_zlib = zlib.decompress(tile1)
                return gzip.compress(tile1_decompressed_zlib)
            else: return  gzip.compress(tile1)   
        
        elif tile2:
            if tile2[:2] == b'\x1f\x8b':
                return tile2
            elif tile2[:2] == b'\x78\x9c' or tile2[:2] == b'\x78\x01' or tile2[:2] == b'\x78\xda':
                tile2_decompressed_zlib = zlib.decompress(tile2)
                return gzip.compress(tile2_decompressed_zlib)     
            else: return  gzip.compress(tile2) 
    except Exception as e:
        return None

def merge_vector_layers(layer1, layer2):
    layer1_ids = {layer['id'] for layer in layer1}
    merged_layers = layer1[:]

    for layer in layer2:
        if layer['id'] in layer1_ids:
            continue  # Skip duplicate layers by 'id'
        merged_layers.append(layer)
    
    return merged_layers

def merge_tilestats(stats1, stats2):
    stats1_layers = {layer['layer']: layer for layer in stats1['layers']}
    merged_layers = stats1['layers'][:]

    for layer in stats2['layers']:
        if layer['layer'] in stats1_layers:
            stats1_layer = stats1_layers[layer['layer']]
            stats1_layer['count'] += layer['count']
            stats1_layer['attributeCount'] += layer['attributeCount']
            stats1_layer['attributes'].extend(layer['attributes'])
        else:
            merged_layers.append(layer)
    
    return {
        "layerCount": len(merged_layers),
        "layers": merged_layers
    }

def merge_json_metadata(json1, json2):
    if not json1:
        return json2
    if not json2:
        return json1

    dict1 = json.loads(json1)
    dict2 = json.loads(json2)

    merged_dict = {**dict1}

    if "vector_layers" in dict1 and "vector_layers" in dict2:
        merged_dict["vector_layers"] = merge_vector_layers(dict1["vector_layers"], dict2["vector_layers"])

    if "tilestats" in dict1 and "tilestats" in dict2:
        merged_dict["tilestats"] = merge_tilestats(dict1["tilestats"], dict2["tilestats"])

    return json.dumps(merged_dict)

def merge_metadata(metadata_dicts):
    merged_metadata = {}
    for metadata in metadata_dicts:
        for name, value in metadata.items():
            if name in merged_metadata:
                if name.strip() =='json':  # Check if the metadata name suggests JSON content
                    merged_metadata[name] = merge_json_metadata(merged_metadata[name], value)
                else:
                    # if merged_metadata[name] != value:
                    merged_metadata[name] += '; ' + value
            else:
                merged_metadata[name] = value
    return merged_metadata

def get_min_zoom(zoom_list):
    try:
        numbers_str = zoom_list.split(';')
        # Convert the number strings to integers or floats
        numbers = [int(num.strip()) for num in numbers_str]
        # Find the minimum value
        min_zoom = min(numbers)
        return str(min_zoom)
    except Exception as e:
        logging.error(f"Get min zoom error: {e}")
        return zoom_list
    
def get_max_zoom(zoom_list):
    try:
        numbers_str = zoom_list.split(';')
        # Convert the number strings to integers or floats
        numbers = [int(num.strip()) for num in numbers_str]
        # Find the minimum value
        max_zoom = max(numbers)
        return str(max_zoom)
    except Exception as e:
        logging.error(f"Get max zoom error: {e}")
        return zoom_list
    
def get_max_bound(bounds_str):
    try:
         # Initialize min and max values with None
        min_lon, min_lat, max_lon, max_lat = None, None, None, None        
        bounds_items = bounds_str.split(';')       
        # Initialize variables to store the min bound
        for item in bounds_items:
            # Extract coordinates from each bound
            coords = list(map(float, item.split(',')))
            if min_lon is None:
                # Initialize min and max with the first bounding box
                min_lon, min_lat, max_lon, max_lat = coords
            else:
                # Update min and max values
                min_lon = min(min_lon, coords[0])
                min_lat = min(min_lat, coords[1])
                max_lon = max(max_lon, coords[2])
                max_lat = max(max_lat, coords[3])
        # Return the max bound as a string
        max_bound = f"{min_lon},{min_lat},{max_lon},{max_lat}"
        return max_bound
    except Exception as e:
        logging.error(f"Get max bound error: {e}")
        return '-180.000000,-85.051129,180.000000,85.051129'

def get_center_of_bound(bounds_str):
    try:
        # Split the string into individual coordinates
        coords = [float(coord) for coord in bounds_str.split(',')]
        if len(coords) == 4:
            lon1, lat1, lon2, lat2 = coords
            # Calculate the center of the bounding box
            center_lon = (lon1 + lon2) / 2
            center_lat = (lat1 + lat2) / 2
            # Return the center as a formatted string
            return f"{center_lon},{center_lat}"
        else:
            raise ValueError("Invalid bounds string format")
    except Exception as e:
        logging.error(f"Get center of bound error: {e}")
        return ''
        
def merge_mbtiles(input_mbtiles, output_mbtiles):
    notexisted_files = [file for file in input_mbtiles if not os.path.exists(file)]

    if notexisted_files:
        print(f"Error: The following input MBTiles files are not existed:")
        for file in notexisted_files:
            print(f"  - {file}")
        return

    if os.path.exists(output_mbtiles):
        os.remove(output_mbtiles)
    is_vector, compression_type = mbtilesfixmeta.check_vector(input_mbtiles[0]) 
    if is_vector:
        mbtilesfixmeta.fix_metadata(input_mbtiles[0], compression_type)   
        shutil.copyfile(input_mbtiles[0], output_mbtiles)
        try:
            conn_out = sqlite3.connect(output_mbtiles)       
            cur_out = conn_out.cursor()
            cur_out.execute('CREATE TABLE IF NOT EXISTS metadata (name TEXT, value TEXT)')
            cur_out.execute('DROP TABLE IF EXISTS tiles_new;')
            cur_out.execute('CREATE TABLE tiles_new (zoom_level INTEGER, tile_column INTEGER, tile_row INTEGER, tile_data BLOB, PRIMARY KEY (zoom_level, tile_column, tile_row))')
            cur_out.execute('''
                INSERT INTO tiles_new (zoom_level, tile_column, tile_row, tile_data)
                SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles;
                ''')
            
            cur_out.execute("SELECT type FROM sqlite_master WHERE name='tiles'")  
            result = cur_out.fetchone()
            if result:
                if result[0] == 'view':
                    cur_out.execute('DROP VIEW tiles')
                elif result[0] == 'table':
                    cur_out.execute('DROP TABLE tiles')
            cur_out.execute('ALTER TABLE tiles_new RENAME TO tiles')
            cur_out.execute('CREATE UNIQUE INDEX tile_index ON tiles (zoom_level, tile_column, tile_row)')
            
            # Merging tiles   
            cur_out.execute('SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles ORDER BY zoom_level')          
            output_rows = cur_out.fetchall()
            tiles = {}

            for (z, x, y, tile) in output_rows:
                key = (z, x, y)
                tiles[key] = tile
            
            connections = [sqlite3.connect(mbtiles) for mbtiles in input_mbtiles]
            cursors_in = [conn.cursor() for conn in connections]
            for i, cursor in enumerate(cursors_in):
                if i == 0:
                    continue  # Skip the cursor for the first MBTiles file     
                is_vector, compression_type = mbtilesfixmeta.check_vector(input_mbtiles[0]) 
                if is_vector:
                    mbtilesfixmeta.fix_metadata(input_mbtiles[i], compression_type)
                    mbtiles_name = os.path.basename(input_mbtiles[i])
                    cursor.execute('SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles ORDER BY zoom_level')
                    rows = cursor.fetchall()
                    for (z, x, y, tile) in tqdm(rows, desc=f"Processing tiles from {mbtiles_name}"):
                        key = (z, x, y)
                        if key in tiles:
                            tiles[key] = merge_tiles(tiles[key], tile, z, x, y)
                        else:
                            tiles[key] = tile

            for key, tile in tqdm(tiles.items(), desc="Inserting merged tiles"):
                cur_out.execute('INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data) VALUES (?, ?, ?, ?)', (key[0], key[1], key[2], tile))
            conn_out.commit()
            print(f"Successfully merged MBTiles files into {output_mbtiles}")
        except Exception as e:
            logging.error(f"Error Merging tile_data {mbtiles_name}: {e}")
        
        try: 
            # Merging metadata
            metadata_dicts = []
            for i, cursor in enumerate(cursors_in):
                mbtiles_name = os.path.basename(input_mbtiles[i])  
                cursor.execute('SELECT name, value FROM metadata')
                metadata_dicts.append({name: value for name, value in cursor.fetchall()})

            merged_metadata = merge_metadata(metadata_dicts)

            for name, value in tqdm(merged_metadata.items(), desc=f"Inserting merged metadata"):
                cur_out.execute('INSERT OR REPLACE INTO metadata (name, value) VALUES (?, ?)', (name, value))

            # Update format
            cur_out.execute(''' UPDATE metadata SET value = 'pbf' where name = 'format' ''')
            
            # Update minzoom
            cur_out.execute("SELECT value FROM metadata WHERE name = 'minzoom'")
            zoom_levels = cur_out.fetchone()[0]  # Fetch the value
            if zoom_levels:
                min_zoom = get_min_zoom(zoom_levels)
                cur_out.execute('''
                    UPDATE metadata 
                    SET value = ? 
                    WHERE name = 'minzoom'
                ''', (min_zoom,))

            # Update maxzoom
            cur_out.execute("SELECT value FROM metadata WHERE name = 'maxzoom'")
            zoom_levels = cur_out.fetchone()[0]  # Fetch the value

            max_zoom = 0
            if zoom_levels:
                max_zoom = get_max_zoom(zoom_levels)
                cur_out.execute('''
                    UPDATE metadata 
                    SET value = ? 
                    WHERE name = 'maxzoom'
                ''', (max_zoom,))
            
            # Update max bounds
            cur_out.execute("SELECT value FROM metadata WHERE name = 'bounds'")
            bounds = cur_out.fetchone()[0]  # Fetch the value
            if bounds:
                min_bound = get_max_bound(bounds)

                cur_out.execute('''
                    UPDATE metadata 
                    SET value = ? 
                    WHERE name = 'bounds'
                    ''', (min_bound,))

            # Update center
            cur_out.execute("SELECT value FROM metadata WHERE name = 'bounds'")
            bound = cur_out.fetchone()[0]  
            center = get_center_of_bound(bound)
            center_of_bound = ''
            if center != '':
                center_of_bound = center +f',{max_zoom}'

            cur_out.execute('''
                UPDATE metadata 
                SET value = ? 
                WHERE name = 'center'
                ''', (center_of_bound,))

            # Update description
            description = 'Merge multiple MBTiles files into a single MBTiles file using mbtilesmerge from vtiles'
            cur_out.execute('''
                INSERT OR REPLACE INTO metadata (name, value)
                VALUES ('description', ?)
            ''', (description,))
        
            conn_out.commit()
            print(f"Successfully merged metadata into {output_mbtiles}")
            
        except Exception as e:
            logging.error(f"Error Merging metadata: {e}")

        finally:
            for conn in connections:
                conn.close()   
            conn_out.close()
    else:
        logging.info('Only vector mbtiles is supported.')
        return

def main():
    parser = argparse.ArgumentParser(description="Merge multiple MBTiles files into a single MBTiles file.")
    parser.add_argument('-i', '--input', nargs='+', required=True, help='Input MBTiles files to merge.')
    parser.add_argument('-o', '--output', required=True, help='Output merged MBTiles file.')

    args = parser.parse_args()

    merge_mbtiles(args.input, args.output)

if __name__ == '__main__':
    main()


import zipfile
import os
import argparse
from tqdm import tqdm

def extract_vtpk(vtpk_path, output_dir):
    # Check if the output directory exists, if not, create it
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Open the VTPK file
    with zipfile.ZipFile(vtpk_path, 'r') as zip_ref:
        # Get the list of files in the archive
        file_list = zip_ref.namelist()
        
        # Use tqdm to display a progress bar while extracting files
        for file in tqdm(file_list, desc="Extracting VTPK", unit="file"):
            zip_ref.extract(file, output_dir)

    print(f'Extracted VTPK to {output_dir}')

def main():
    # Initialize the argument parser
    parser = argparse.ArgumentParser(description="Extract VTPK (Vector Tile Package) to a specified folder.")
    
    # Define input and output arguments
    parser.add_argument('-i', '--input', required=True, help="Path to the input VTPK file")
    parser.add_argument('-o', '--output', required=True, help="Path to the output directory")
    
    # Parse the command-line arguments
    args = parser.parse_args()

    # Extract the VTPK
    extract_vtpk(args.input, args.output)

if __name__ == '__main__':
    main()



import pyclipper
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry.polygon import Polygon
from shapely.ops import unary_union
from shapely.validation import explain_validity


def _coords(shape):
    """
    Return a list of lists of coordinates of the polygon. The list consists firstly of the list of exterior
    coordinates followed by zero or more lists of any interior coordinates.
    """

    assert shape.geom_type == "Polygon"
    coords = [list(shape.exterior.coords)]
    for interior in shape.interiors:
        coords.append(list(interior.coords))
    return coords


def _drop_degenerate_inners(shape):
    """
    Drop degenerate (zero-size) inners from the polygon.

    This is implemented as dropping anything with a size less than 0.5, as the polygon is in integer coordinates and
    the smallest valid inner would be a triangle with height and width 1.
    """
    assert shape.geom_type == "Polygon"

    new_inners = []
    for inner in shape.interiors:
        # need to make a polygon of the linearring to get the _filled_ area of
        # the closed ring.
        if abs(Polygon(inner).area) >= 0.5:
            new_inners.append(inner)

    return Polygon(shape.exterior, new_inners)


def _contour_to_poly(contour):
    poly = Polygon(contour)
    if not poly.is_valid:
        poly = poly.buffer(0)
    assert poly.is_valid, f"Contour {contour!r} did not make valid polygon {poly.wkt} because {explain_validity(poly)}"
    return poly


def _union_in_blocks(contours, block_size):
    """
    Generator which yields a valid shape for each block_size multiple of input contours. This merges together the
    contours for each block before yielding them.
    """
    n_contours = len(contours)
    for i in range(0, n_contours, block_size):
        j = min(i + block_size, n_contours)

        inners = []
        for c in contours[i:j]:
            p = _contour_to_poly(c)
            if p.geom_type == "Polygon":
                inners.append(p)
            elif p.geom_type == "MultiPolygon":
                inners.extend(p.geoms)
        holes = unary_union(inners)
        assert holes.is_valid

        yield holes


def _generate_polys(contours):
    """
    Generator which yields a valid polygon for each contour input.
    """
    for c in contours:
        p = _contour_to_poly(c)
        yield p


def _polytree_node_to_shapely(node):
    """
    Recurses down a Clipper PolyTree, extracting the results as Shapely objects.

    Returns a tuple of (list of polygons, list of children)
    """
    polygons = []
    children = []
    for ch in node.Childs:
        p, c = _polytree_node_to_shapely(ch)
        polygons.extend(p)
        children.extend(c)

    if node.IsHole:
        # Check expectations: a node should be a hole, _or_ return children. This is because children of holes must
        # be outers, and should be on the polygons list.
        assert len(children) == 0
        children = [node.Contour] if node.Contour else []

    elif node.Contour:
        poly = _contour_to_poly(node.Contour)

        # We add each inner one-by-one so that we can reject them individually if they cause the polygon to become
        # invalid. If the shape has lots of inners, then this can mean a proportional amount of work, and may take
        # 1,000s of seconds. Instead, we can group inners together, which reduces the number of times we call the
        # expensive 'difference' method.
        block_size = 200
        inners = _union_in_blocks(children, block_size) if len(children) > block_size else _generate_polys(children)

        for inner in inners:
            # The difference of two valid polygons may fail, and in this  situation we'd like to be able to display
            # the polygon anyway. So we discard the bad inner and continue.
            #
            # See test_polygon_inners_crossing_outer for a test case.
            try:
                diff = poly.difference(inner)
            except Exception:
                continue

            if not diff.is_valid:
                diff = diff.buffer(0)

            # Keep this for when https://trac.osgeo.org/geos/ticket/789 is resolved.
            #
            # assert diff.is_valid, (
            #     f"Difference of {poly.wkt} and {inner.wkt} did not make valid polygon {diff.wkt} "
            #     f" because {explain_validity(diff)}"
            # )
            # NOTE: this throws away the inner ring if we can't produce a valid difference. Not ideal, but we'd
            # rather produce something that's valid than nothing.
            if diff.is_valid:
                poly = diff

        assert poly.is_valid
        if poly.geom_type == "MultiPolygon":
            polygons.extend(poly.geoms)
        else:
            polygons.append(poly)
        children = []

    else:
        # Check expectations: this branch gets executed if this node is not a hole, and has no contour. In that
        # situation we'd expect that it has no children, as it would not be possible to subtract children from
        # an empty outer contour.
        assert len(children) == 0

    return polygons, children


def _polytree_to_shapely(tree):
    polygons, children = _polytree_node_to_shapely(tree)

    # expect no left over children - should all be incorporated into polygons by the time recursion returns to the root.
    assert len(children) == 0

    union = unary_union(polygons)
    assert union.is_valid
    return union


def make_valid_pyclipper(shape):
    """
    Use the pyclipper library to "union" a polygon on its own. This operation uses the even-odd rule to determine
    which points are in the interior of the polygon, and can reconstruct the orientation of the polygon from that.
    The pyclipper library is robust, and uses integer coordinates, so should not produce any additional degeneracies.

    Before cleaning the polygon, we remove all degenerate inners. This is useful to remove inners which have
    collapsed to points or lines, which can interfere with the cleaning process.
    """

    # drop all degenerate inners
    clean_shape = _drop_degenerate_inners(shape)

    pc = pyclipper.Pyclipper()

    try:
        pc.AddPaths(_coords(clean_shape), pyclipper.PT_SUBJECT, True)

        # note: Execute2 returns the polygon tree, not the list of paths
        result = pc.Execute2(pyclipper.CT_UNION, pyclipper.PFT_EVENODD)

    except pyclipper.ClipperException:
        return MultiPolygon([])

    return _polytree_to_shapely(result)


def make_valid_polygon(shape):
    """
    Make a polygon valid. Polygons can be invalid in many ways, such as self-intersection, self-touching and
    degeneracy. This process attempts to make a polygon valid while retaining as much of its extent or area as possible.

    First, we call pyclipper to robustly union the polygon. Using this on its own appears to be good for "cleaning"
    the polygon.

    This might result in polygons which still have degeneracies according to the OCG standard of validity - as
    pyclipper does not consider these to be invalid. Therefore, we follow by using the `buffer(0)` technique to
    attempt to remove any remaining degeneracies.
    """
    assert shape.geom_type == "Polygon"

    shape = make_valid_pyclipper(shape)
    assert shape.is_valid
    return shape


def make_valid_multipolygon(shape):
    new_g = []

    for g in shape.geoms:
        if g.is_empty:
            continue

        valid_g = make_valid_polygon(g)

        if valid_g.geom_type == "MultiPolygon":
            new_g.extend(valid_g.geoms)
        else:
            new_g.append(valid_g)

    return MultiPolygon(new_g)


def make_it_valid(shape):
    """
    Attempt to make any polygon or multipolygon valid.
    """
    if shape.is_empty:
        return shape

    elif shape.geom_type == "MultiPolygon":
        shape = make_valid_multipolygon(shape)

    elif shape.geom_type == "Polygon":
        shape = make_valid_polygon(shape)

    return shape



import warnings

from . import decoder, encoder


def decode(tile, per_layer_options=None, default_options=None, **kwargs):
    """Decode the provided `tile`

    Args:
        tile:
            The tile to decode.

        per_layer_options:
            An optional dictionary containing per layer options. The keys are the layer names and the values are
            options as described further. If an option is missing for a layer or if a layer is missing, the values of
            `default_options` are taken first. Finally, the global default options are used.

        default_options:
            These options are taken for layers without entry in `per_layer_options`. For all missing options values,
            the global default values are taken.

    Returns:
        The decoded layers data.

    Notes:
        The possible options are:
            * `y_coord_down`: it suppresses flipping the y coordinate values during encoding when set to `True`.
            Default to `False`.
            * `transformer`: a function transforming the coordinates of geometry object. It takes two floats (`x`
            and `y`) as arguments and retrieves the transformed coordinates `x_transformed`, `y_transformed`. Default to
            `None`.
            * `geojson`: when set to `False`, the behaviour of mapbox-vector-tile version 1.* is used. When set
            to `False`, the retrieved dictionary is a valid geojson file. Default to `True`.
    """
    if kwargs:
        warnings.warn("`decode` signature has changed, use `default_options` instead", DeprecationWarning, stacklevel=2)
        default_options = {**kwargs, **(default_options or {})}
    vector_tile = decoder.TileData(pbf_data=tile, per_layer_options=per_layer_options, default_options=default_options)
    message = vector_tile.get_message()
    return message


def encode(layers, per_layer_options=None, default_options=None, **kwargs):
    """Encode the `layers` into a MVT tile.

    Args:
        layers:
            The layer data to encode.

        per_layer_options:
            An optional dictionary containing per layer options. The keys are the layer names and the values are
            options as described further. If an option is missing for a layer or if a layer is missing, the values of
            `default_options` are taken first. Finally, the global default options are used.

        default_options:
            These options are taken for layers without entry in `per_layer_options`. For all missing options values,
            the global default values are taken.

    Returns:
        The encoded tile.

    Notes:
        The possible options are:
            * `y_coord_down`: it suppresses flipping the y coordinate values during encoding when set to `True`.
            Default to `False`.
            * `transformer`: a function transforming the coordinates of geometry object. It takes two floats (`x`
            and `y`) as arguments and retrieves the transformed coordinates `x_transformed`, `y_transformed`. Default to
            `None`.
            * `quantize_bounds`: bounds in the form (minx, miny, maxx, maxy) used to scale coordinates during
            encoding. Default to `None`.
            * `extents`: extents of the tile which is passed through to the layer in the pbf, and honored during any
            quantization or y coordinate flipping. Default to 4096.
            * `on_invalid_geometry`: a function taking a shapely shape as argument and retrieving an optional
            valid geometry. Default to None. In the file `encoder.py`, three possible functions are defined:
                * `on_invalid_geometry_raise`: it raises an error if an invalid geometry exists.
                * `on_invalid_geometry_ignore`: it ignores the invalid geometry and replaces it with a `None`.
                * `on_invalid_geometry_make_valid`: it tries to make the geometry valid. If it fails, retrieves `None`.
            * `check_winding_order`: it forces the check of the winding order for polygons. Default to True.
            * `max_geometry_validate_tries`: the number of tries when trying to enforce the good winding order. Default
            to 5.
    """
    if kwargs:
        warnings.warn("`encode` signature has changed, use `default_options` instead", DeprecationWarning, stacklevel=2)
        default_options = {**kwargs, **(default_options or {})}
    vector_tile = encoder.VectorTile(default_options=default_options)
    if per_layer_options is None:
        per_layer_options = {}
    if isinstance(layers, list):
        for layer in layers:
            layer_name = layer["name"]
            layer_options = per_layer_options.get(layer_name, None)
            vector_tile.add_layer(features=layer["features"], name=layer_name, options=layer_options)
    else:
        layer_name = layers["name"]
        layer_options = per_layer_options.get(layer_name, None)
        vector_tile.add_layer(features=layers["features"], name=layer_name, options=layer_options)

    return vector_tile.tile.SerializeToString()



import json
import mmap
from .tile import (
    deserialize_header,
    deserialize_directory,
    zxy_to_tileid,
    tileid_to_zxy,
    find_tile,
    Compression,
)
import gzip


def MmapSource(f):
    mapping = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)

    def get_bytes(offset, length):
        return mapping[offset : offset + length]

    return get_bytes


def MemorySource(buf):
    def get_bytes(offset, length):
        return buf[offset : offset + length]

    return get_bytes


class Reader:
    def __init__(self, get_bytes):
        self.get_bytes = get_bytes

    def header(self):
        return deserialize_header(self.get_bytes(0, 127))

    def metadata(self):
        header = self.header()
        metadata = self.get_bytes(header["metadata_offset"], header["metadata_length"])
        if header["internal_compression"] == Compression.GZIP:
            metadata = gzip.decompress(metadata)
        return json.loads(metadata)

    def get(self, z, x, y):
        tile_id = zxy_to_tileid(z, x, y)
        header = self.header()
        dir_offset = header["root_offset"]
        dir_length = header["root_length"]
        for depth in range(0, 4):  # max depth
            directory = deserialize_directory(self.get_bytes(dir_offset, dir_length))
            result = find_tile(directory, tile_id)
            if result:
                if result.run_length == 0:
                    dir_offset = header["leaf_directory_offset"] + result.offset
                    dir_length = result.length
                else:
                    return self.get_bytes(
                        header["tile_data_offset"] + result.offset, result.length
                    )


def traverse(get_bytes, header, dir_offset, dir_length):
    entries = deserialize_directory(get_bytes(dir_offset, dir_length))
    for entry in entries:
        if entry.run_length > 0:
            for i in range(entry.run_length):
                yield tileid_to_zxy(entry.tile_id + i), get_bytes(
                    header["tile_data_offset"] + entry.offset, entry.length
                )
        else:
            for t in traverse(
                get_bytes,
                header,
                header["leaf_directory_offset"] + entry.offset,
                entry.length,
            ):
                yield t


def all_tiles(get_bytes):
    header = deserialize_header(get_bytes(0, 127))
    return traverse(get_bytes, header, header["root_offset"], header["root_length"])



import os
import json


def current_dir(file):
    return os.path.dirname(os.path.abspath(file))


def get_parent_dir(directory):
    return os.path.abspath(os.path.join(directory, os.path.pardir))


def get_json(file_path):
    data = None
    with open(file_path) as json_file:
        data = json.load(json_file)
        _change_int_coords_to_float(data)
    return data


def _change_int_coords_to_float(data):
    if isinstance(data, dict):
        _walk_dict(data)
    if isinstance(data, list):
        _walk_list(data)


def _walk_dict(tree):
    for key, value in tree.items():
        if isinstance(value, dict):
            _walk_dict(value)
        if isinstance(value, list):
            _walk_list(value)


def _walk_list(lst):
    if len(lst) == 0:
        return
    if isinstance(lst[0], list):
        for l in lst:
            _walk_list(l)
    elif isinstance(lst[0], int):
        for i in range(len(lst)):
            lst[i] = float(lst[i])
    elif isinstance(lst[0], dict):
        for i in range(len(lst)):
            _walk_dict(lst[i])



import os, math
import sqlite3
import json
import argparse
from tqdm import tqdm
from vtiles.utils.geopreocessing import flip_y, safe_makedir

def extract_metadata(cursor):
  """Extract metadata from MBTiles file."""
  cursor.execute("SELECT name, value FROM metadata")
  metadata_rows = cursor.fetchall()
  metadata = {}
  for name, value in metadata_rows:
    metadata[name] = value
  return metadata

def write_metadata_to_json(metadata, dirname):
  """Write metadata to JSON file."""
  metadata_json_path = os.path.join(dirname, "metadata.json")
  with open(metadata_json_path, "w") as metadata_file:
      json.dump(metadata, metadata_file, indent=4)
  print("Writing metadata.json done!")

def determine_tile_format(cursor):
  """Determine tile format based on metadata."""
  cursor.execute("SELECT value FROM metadata WHERE name='format'")
  tile_format = cursor.fetchone()[0]
  if tile_format:
      if tile_format== 'png' or tile_format == 'webp':
          return 'png'
      elif tile_format == 'jpg':
          return 'jpg'
      elif tile_format == 'pbf':
          return 'pbf'
  return ''

def get_max_zoom(cursor):
    cursor.execute('SELECT MAX(zoom_level) FROM tiles')
    max_zoom = cursor.fetchone()[0]
    return max_zoom

def convert_mbtiles_to_folder(input_mbtiles, output_folder, flipy=0, min_zoom=0, max_zoom=None):
    connection = sqlite3.connect(input_mbtiles)
    cursor = connection.cursor()
    
    safe_makedir(output_folder)
    tile_format = determine_tile_format(cursor)
    
    mbtiles_max_zoom = get_max_zoom(cursor)
    if max_zoom is None or max_zoom > mbtiles_max_zoom:
        max_zoom = mbtiles_max_zoom

   
    metadata = extract_metadata(cursor)
    write_metadata_to_json(metadata, output_folder)
    
    cursor.execute('SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles WHERE zoom_level BETWEEN ? AND ? ORDER BY zoom_level', (min_zoom, max_zoom))
    tiles = cursor.fetchall()

    for zoom, col, row, tile_data in tqdm(tiles, unit = ' tiles ', desc='Processing tiles'):
        # Flip the Y coordinate if flipy is True
        y = flip_y(zoom, row) if flipy else row

        # Construct the directory path
        tile_dir = os.path.join(output_folder, str(zoom), str(col))

        # Ensure the directory exists
        safe_makedir(tile_dir)

        # Construct the file path
        tile_path = os.path.join(tile_dir, f'{y}.{tile_format}')

        # Write the tile data to the file
        with open(tile_path, 'wb') as tile_file:
            tile_file.write(tile_data)

    print('Converting mbtiles to folder done!')
    connection.close()

def main():
    parser = argparse.ArgumentParser(description='Convert MBTiles file to folder')
    parser.add_argument('-i', required=True, help='Input MBTiles file name')
    parser.add_argument('-o', help='Output folder name (optional)')
    parser.add_argument('-flipy',type=int, default=0, choices=[0, 1], help='TMS <--> XYZ tiling scheme (optional): 1 or 0, default is 0')
    parser.add_argument('-minzoom', type=int, default=0, help='Min zoom to export (optional, default is 0)')
    parser.add_argument('-maxzoom', type=int, default=None, help='Max zoom to export (optional, default is maxzoom from the input MBTiles)')

    args = parser.parse_args()

    if not os.path.exists(args.i):
        print('MBTiles file does not exist! Please recheck and input a correct file path.')
        exit()
    
    input_filename_abspath = os.path.abspath(args.i)
    output_folder_abspath = os.path.abspath(args.o) if args.o else os.path.join(os.path.dirname(args.i), os.path.splitext(os.path.basename(args.i))[0])
    
    print(f'Converting {input_filename_abspath} to {output_folder_abspath} folder.')
    convert_mbtiles_to_folder(args.i, output_folder_abspath, args.flipy, args.minzoom, args.maxzoom)

if __name__ == "__main__":
    main()



import argparse,os
import sqlite3
import zlib
import gzip
import logging
import shutil
from tqdm import tqdm
from vtiles.mbtiles import mbtilesfixmeta, mbtilesdelduplicate

def decompress_tile_data(tile_data):
    try:
        if tile_data[:2] == b'\x1f\x8b':  # Check for gzip magic number
            tile_data = gzip.decompress(tile_data)
        elif tile_data[:2] == b'\x78\x9c' or tile_data[:2] == b'\x78\x01' or tile_data[:2] == b'\x78\xda':
            tile_data = zlib.decompress(tile_data) 
    except Exception as e:
        logging.error(f"Failed to decompress tile data: {e}")
        return tile_data
    return tile_data          

def decompress_mbtiles(input_mbtiles, output_mbtiles):
    mbtilesdelduplicate.remove_duplicates(input_mbtiles,output_mbtiles)

    # Open the copied MBTiles file
    conn = sqlite3.connect(output_mbtiles)
    cursor = conn.cursor()

    # Check if the tiles table is a view
    cursor.execute("SELECT type FROM sqlite_master WHERE name='tiles'")
    result = cursor.fetchone()

    if result and result[0] == 'view':
        # Create a new table named tiles_new
        cursor.execute("CREATE TABLE tiles_new (zoom_level INTEGER, tile_column INTEGER, tile_row INTEGER, tile_data BLOB)")
        
        # Select data from tiles view
        cursor.execute("SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles order by zoom_level")
        rows = cursor.fetchall()
        
        # Insert decompressed data into the new table
        for zoom_level, tile_column, tile_row, tile_data in tqdm(rows, desc="Decompressing tiles", unit="tile"):
            try:
                decompressed_data = decompress_tile_data(tile_data)
                cursor.execute(
                    "INSERT INTO tiles_new (zoom_level, tile_column, tile_row, tile_data) VALUES (?, ?, ?, ?)",
                    (zoom_level, tile_column, tile_row, decompressed_data)
                )
            except Exception as e:
                print(f"Error decompressing tile {zoom_level}/{tile_column}/{tile_row}: {e}")
            
        # Drop the view and rename the new table to tiles
        cursor.execute("DROP VIEW tiles")
        cursor.execute("ALTER TABLE tiles_new RENAME TO tiles")
        cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS tile_index ON tiles (zoom_level, tile_column, tile_row)")
    else:
        # Decompress tile data in the existing tiles table
        cursor.execute("SELECT zoom_level, tile_column, tile_row, tile_data FROM tiles order by zoom_level")
        tiles = cursor.fetchall()
        
        # Add tqdm progress bar
        for zoom_level, tile_column, tile_row, tile_data in tqdm(tiles, desc="Decompressing tiles", unit="tile"):
            try:
                decompressed_data = decompress_tile_data(tile_data)
                cursor.execute(
                    "UPDATE tiles SET tile_data = ? WHERE zoom_level = ? AND tile_column = ? AND tile_row = ?",
                    (decompressed_data, zoom_level, tile_column, tile_row)
                )
            except Exception as e:
                print(f"Error decompressing tile {zoom_level}/{tile_column}/{tile_row}: {e}")
        
        cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS tile_index ON tiles (zoom_level, tile_column, tile_row)")
    # Commit and close connections
    conn.commit()
    conn.close()


def main():
    parser = argparse.ArgumentParser(description='Decompress an MBTiles file.')
    parser.add_argument('-i', '--input', required=True, help='Path to the input MBTiles file.')
    parser.add_argument('-o', '--output', required=True, help='Path to the output MBTiles file.')

    args = parser.parse_args()
    decompress_mbtiles(args.input, args.output)

if __name__ == "__main__":
    main()
